from view import agnostic as                    _a 
from view import grammar as                     _g
from view import layer as                       _l
from view import py_form as                     _pf
from view import py_rule as                     _pr
import                                          random
import rhinoscriptsyntax as                     _rs
from view import settings as                    _s
import sortalgi as                              _sgi
import                                          sortal
from view import view_main as                   _vm
from view import view_utilities as              _vu

origin = _s.Settings.origin

class ControllerMain(object):
    """Talks to Sortal and ViewMain only, not Rhino

    shape: An SG shape
    form: A set of guids of curves and text dots

    apply_rule(a, b, c)
    add_shapes(a, b)
    subtract_shapes(a, b)

    Three representations of a shape:
    1. Shape grammar: Shape
    2. Rhino: Form (has a list of object guids)
    3. Agnostic / neutral / controller: list of specs
    """
    class_name = 'ControllerMain'

    def __init__(self):
        self.vm = _vm.ViewMain()

    @classmethod
    def reset_sortal(cls):
        sortal.rule.cleanup()
        value = _sgi.sortal_setup()
        if value:
            message_1 = "Sortal has been initialized"
        else:
            message_1 = "Sortal has not been initialized"
        print(message_1)
        message_2 = "Initialize the Rhino document also? [y or n]"
        clear_rhino_doc = _rs.GetString(message_2)
        if clear_rhino_doc in ['y', 'Y']:
            _vm.ViewMain.clear_all()
            _l.Layer.new('Shape 0')
            message_3a = ('%s %s' % (
                "Draw the initial shape in the +/+ quadrant", 
                "on the layer 'Shape 0'."))
            message_3b = ('%s %s' % (
                "Draw everything else in the -/+ and -/- quadrants", 
                "on the layer 'Default'"))
            message_3 = ('%s %s' % (message_3a, message_3b))
            print(message_3)

    def load_rules(self):
        """Precondition:
            Sortal register is empty
        From rhino rules creates sortal rules ('loads' rules into the sortal 
        register. Returns:
            n               int. The number of rules loaded successfully
        """
        rule_layer_names = self.vm.get_rule_layer_names()
        py_rules = _pr.PyRule.many_new_from_rule_groups(rule_layer_names)
                                                ##  move to ViewMain?
        n = self.create_sortal_rules(py_rules, rule_layer_names)
        message = '%i rules loaded' % n
        print(message)
        return n

    def create_rule(self):
        """Prompts the user to create a rule from existing Rhino objects. 
        Draws the new rule on a new layer. Creates a sortal rule. Rearranges 
        the existing rules. Returns:
            new_rh_rule     RhRule. The new RhRule object. None if 
                            unsuccessful
        """
        prr = py_rule_received = self.vm.request_new_rule()
        if prr:
            if self.vm.rh_rule_exists(prr):
                message_rule_exists = '%s %s' % (
                    'The rule already exists.', 
                    'Create another copy? [y/n]')
                response = _rs.GetString(message_rule_exists)
                yes = response in ['y', 'yes', 'Y', 'Yes']
                no = response in ['n', 'no', 'No', 'No']
                if yes:
                    new_rh_rule = prr.draw()
                    n_rule_obs_moved = self.vm.arrange_rh_rules(new_rh_rule)

                    ag_left = prr.a.ag
                    ag_right = prr.b.ag
                    name = new_rh_rule.layer_name
                    descr = prr.descr
                    self.create_sortal_rule(ag_left, ag_right, name, descr)
                    return_value = new_rh_rule
                elif no:
                    message_did_not_duplicate = (
                        'Did not duplicate the existing rule')
                    print(message_did_not_duplicate)
                    return_value = None
                else:
                    error_message = "Shouldn't be here"
                    print(error_message)
                    return_value = None
            else:
                new_rh_rule = prr.draw()
                n_rule_obs_moved = self.vm.arrange_rh_rules(new_rh_rule)
                ag_left = prr.a.ag
                ag_right = prr.b.ag
                name = new_rh_rule.layer_name
                descr = prr.descr
                self.create_sortal_rule(ag_left, ag_right, name, descr)
                return_value = new_rh_rule
        elif prr == None:
            message_none = 'Ill-formed rule. Maybe the left shape was empty?'
            print(message_none)
            return_value = None
        else:
            error_message = "Shouldn't be here"
            print(error_message)
            return_value = None
        return return_value

    def get_rule_draw_next_shapes(self):
        """Prompts the user to select the rule application components 
        (current form, current subform, rule). Finds and draws all distinct 
        next forms in new layers if successful. Shows message otherwise. 
        Returns:
            n               int. The number of forms drawn. None if 
                            unsuccessful
        """
        if _g.Grammar.is_empty():               ##  move to ViewMain?
            message = "The grammar is empty: nothing to do"
            print(message)
            return_value = 0
        else:
            form_result = self.vm.request_form_and_subform()
            if form_result == None:
                return_value = 0
            else:
                py_form, sub_py_form = form_result
                rule_result = self.vm.request_rule()
                if type(rule_result) == str:
                    rule_name = rule_result
                    next_result = self.apply_rule_by_name(
                        py_form, sub_py_form, rule_name)
                    if (type(next_result) == list and
                        not next_result == []
                    ):
                        next_py_forms = next_result
                        forms_shifted = _pf.PyForm.draw_many_nicely(
                            next_py_forms)      ##  move to ViewMain?
                        n = len(next_py_forms)
                        return_value = n
                    else:
                        return_value = not_a_non_empty_list = None
                elif rule_result == None:
                    return_value = cancelled = 0
                else:
                    return_value = trouble = None
        return return_value

    def get_rule_draw_one_next_shape_repeat(self):
    #                                           ##  2018-12-26 19:30
        """Prompts the user to select a rule, a current form, and the 
        maximum number of applications. Applies the rule to the form. On a new 
        layer, draws one randomly selected next form, if any. Repeats up to 
        the maximum; then prompts the user whether to continue. Stops if 
        there is no next form. Returns:
            final_pf        PyForm. The final py_form, if any
        """
        if _g.Grammar.is_empty():
            return_value = None
        else:
            message_curr_pf = "Select the current shape"
            curr_pf = self.vm.request_non_empty_form(message_curr_pf)

            error_message_rule = '%s %s' % (
                "Don't create a rule now.", 
                "Select an existing rule"
            )
            rule_result = self.vm.request_rule()
            while not type(rule_result) == str:
                print(error_message_rule)
                rule_result = self.vm.request_rule()
            layer_name = rule_result

            message_n = "How many times to repeat the rule application? [3]"
            n = _rs.GetInteger(message_n)
            if n:
                final_pf = self.apply_rule_draw_one_next_shape_repeat(
                    curr_pf, layer_name, n)
            else:
                final_pf = self.apply_rule_draw_one_next_shape_repeat(
                    curr_pf, layer_name)
            return_value = final_pf
        return return_value

    def apply_rule_draw_one_next_shape_repeat(self, curr_pf, layer_name, n=3):
        """Receives:
            curr_pf         PyForm. The current PyForm
            layer_name      str. The name of the layer containing the rule
            n               int. Maximum number of repetitions
        Applies the named rule, selects one next form, if any. Repeats up to 
        n times. Stops when n is reached or if there is no next form. 
        Returns: 
            final_pf        PyForm. The final next PyForm. None if none or if 
                            unsuccessful
        """
        for i in range(n):
            next_pf = self.apply_rule_draw_one_next_shape(
                curr_pf, layer_name)
            if not next_pf:
                if i == 0:
                    final_pf = None
                else:
                    final_pf = curr_pf
                break
            else:
                curr_pf = next_pf
            final_pf = next_pf
        return final_pf

    def apply_rule_draw_one_next_shape(self, curr_pf, layer_name):
        """Receives:
            curr_pf         PyForm. The current PyForm
            layer_name      str. The name of the layer containing the rule
        Returns:
            next_pf         PyForm. The next PyForm, if any. None if none or 
                            unsuccessful
        """
        if _g.Grammar.is_empty():
            message = "The grammar is empty. Create a shape and a rule"
            print(message)
            return_value = None
        elif not type(curr_pf) == _pf.PyForm:
            message = "The current shape must be a PyForm"
            print(message)
            return_value = None
        elif not type(layer_name) == str:
            message = "The layer name must be a string"
            print(message)
            return_value = None
        elif not _rs.IsLayer(layer_name):
            message = "The named layer must exist"
            print(message)
            return_value = None
        elif not curr_pf.is_well_formed:
            message = "The current shape is ill-formed"
            print(message)
            return_value = None
        elif not self.is_a_sortal_rule(layer_name):
            message = "The sortal engine does not contain this rule"
            print(message)
            return_value = None
        else:
            curr_sub_pf = curr_pf
            next_pfs = next_py_forms = self.apply_rule_by_name(
                curr_pf, curr_sub_pf, layer_name)
            next_pf = self._select_random(next_pfs)
            if next_pf:
                _pf.PyForm.draw_many_nicely([next_pf])
                return_value = next_pf
            else:
                return_value = None
        return return_value

    def _select_random(self, items):
        """Receives:
            items           [item]. A list of items of any type
        Returns:
            item            item. A random item in the list. None if the list 
                            is empty
        """
        n = len(items)
        if n == 0:
            item = None
        elif n == 1:
            item = items[0]
        else:
            i = random.randint(0, n - 1)
            item = items[i]
        return item

    def apply_rule_by_name(self, curr_py_form, curr_sub_py_form, layer_name):
        """Receives:
            curr_py_form    PyForm
            curr_sub_py_form
                            PyForm
            layer_name      str
        Applies the rule to the current form and subform. Returns:
            next_py_forms   [PyForm, ...] if successful. None otherwise
        """
        method_name = 'apply_rule_by_name'
        if not (
            type(curr_py_form) == _pf.PyForm and
            type(curr_sub_py_form) == _pf.PyForm and
            type(layer_name) == str
        ):
            message = "The arguments must be 2 PyForms and an integer"
            _vu._print_error_message(self.class_name, method_name, message)
            return_value = None
        elif _l.Layer._is_available(layer_name):
            message = "This rule does not exist"
            _vu._print_error_message(self.class_name, method_name, message)
            return_value = None
        elif not self.is_a_sortal_rule(layer_name):
            message = "The sortal register does not contain this rule"
            _vu._print_error_message(self.class_name, method_name, message)
            return_value = None
        else:
            ag_shape = curr_py_form.ag
            sub_ag_shape = curr_sub_py_form.ag  ##  check that d <= c
            next_appns = _sgi.find_rule_appns(
                layer_name, ag_shape, sub_ag_shape)
            next_ag_shapes = ControllerMain._extract_ag_shapes_from_appns(
                next_appns)
            next_py_forms = _pf.PyForm.some_new_from_agnostics(next_ag_shapes)
            return_value = next_py_forms
        return return_value

    def apply_py_rule(
        self, curr_py_form, curr_sub_py_form, py_rule, rule_name
    ):
        """Receives:
            curr_py_form    PyForm. The current PyForm. Must not be empty. 
                            (Unless sortal takes care of it)
            curr_sub_py_form
                            PyForm. The current sub-PyForm
            py_rule         PyRule. The PyRule to be applied
            rule_name       str. The name of the rule and layer
        Creates a sortal rule. Applies the rule to the current form and 
        subform. If c !<= d, receives a message box from sortal. Returns:
            next_py_forms   [PyForm, ...] if acceptable to sortal. None 
                            otherwise
        """
        if not (
            type(curr_py_form) == _pf.PyForm and
            type(curr_sub_py_form) == _pf.PyForm and
            type(py_rule) == _pr.PyRule and
            type(rule_name) == str
        ):
            message = "The arguments must be 2 PyForms, a PyRule, and a str"
            print(message)
            return_value = None
        else:
            ag_shape = curr_py_form.ag
            sub_ag_shape = curr_sub_py_form.ag  ##  check that d <= c?
            rule_descr = py_rule.descr
            ag_left = py_rule.a.ag
            ag_right = py_rule.b.ag
            sortal_rule = self.create_sortal_rule(
                ag_left, ag_right, rule_name, rule_descr)
            next_appns = _sgi.find_rule_appns(
                sortal_rule, ag_shape, sub_ag_shape)
            if next_appns:
                next_ag_shapes = ControllerMain._extract_ag_shapes_from_appns(
                    next_appns)
                next_py_forms = _pf.PyForm.some_new_from_agnostics(
                    next_ag_shapes)
                return_value = next_py_forms
            else:
                return_value = next_appns
        return return_value

    @classmethod
    def _extract_ag_shapes_from_appns(cls, next_appns):
        """Receives:
            next_appns      [appn, ...]. A list of rule applications 
                appn        (t, t_of_a_in_d, t_of_b, ag_next)
                t           [matrix_row, matrix_row, matrix_row]. A 
                            transformation matrix
                    matrix_row
                            [num, num, num]
                t_of_a_in_d ag_shape
                t_of_b      ag_shape
                ag_next     ag_shape
        Returns:
            next_ag_shapes  [ag_shape, ...]
        """
        next_ag_shapes = []
        for appn in next_appns:
            t, t_of_a_in_d, t_of_b, ag_next = appn
            next_ag_shapes.append(ag_next)
        return next_ag_shapes

    @classmethod
    def create_sortal_rules(cls, py_rules, names_in):
        """Receives:
            py_rules        [PyRule, ...]
            names_in        [str, ...]. len(py_rule_names) == len(py_rules)
        Returns:
            n               int. The number of sortal rules created
        """
        method_name = 'create_sortal_rules'
        n_py_rules = len(py_rules)
        n_names_in = len(names_in)
        n = n_sortal_rules = 0
        if not n_py_rules == n_names_in:
            pass
        else:
            for i in range(n_py_rules):
                pr = py_rules[i]
                ag_left = pr.a.ag
                ag_right = pr.b.ag
                name = names_in[i]
                descr = pr.descr
                sortal_rule = cls.create_sortal_rule(
                    ag_left, ag_right, name, descr)
                if sortal_rule:
                    n += 1
        return n

    @classmethod
    def create_sortal_rule(cls, ag_left, ag_right, name, descr=''):
        """Receives:
            ag_left         ag_shape. Must not be empty
            ag_right        ag_shape
            name            str. Guaranteed unique among layer names. 
            descr           str [or num]
        Wraps _sgi.create_rule. Does not overwrite an existing sortal rule. 
        Returns:
            return_value    sortal_rule. If successful. None otherwise, 
                            including if the name is in use in the sortal 
                            register

        Idea: Success: the rule is registered without overwriting another 
        rule 
        """
        method_name = 'create_sortal_rule'
        if not type(ag_left) == dict:
            message = '%s.%s: Bad type: ag_left' % (
                cls.class_name, method_name)
            print(message)
            return_value = None
        elif not type(ag_right) == dict:
            message = '%s.%s: Bad type: ag_right' % (
                cls.class_name, method_name)
            print(message)
            return_value = None
        elif not type(name) == str:
            message = '%s.%s: Bad type: name' % (cls.class_name, method_name)
            print(message)
            return_value = None
        elif not type(descr) == str:
            message = '%s.%s: Bad type: descr' % (cls.class_name, method_name)
            print(message)
            return_value = None
        elif _a.Agnostic.is_empty_ag_shape(ag_left): ##
            message = '%s.%s: Bad value: ag_left' % (
                cls.class_name, method_name)
            print(message)
            return_value = None
        elif ControllerMain.is_a_sortal_rule(name):
            message = '%s.%s: Bad value: name. Already in use by sortal' % (
                cls.class_name, method_name)
            print(message)
            return_value = None
        else:
            sortal_rule, sortal_rule_name = _sgi.create_rule(
                name, descr, ag_left, ag_right)
            if sortal_rule_name == name:
                return_value = sortal_rule
                message = 'Sortal rule %s created' % sortal_rule_name
                print(message)
            else:
                message = "Received the wrong name"
                print('%s.%s: %s' % (cls.class_name, method_name, message))
                return_value = None
        return return_value

    @classmethod
    def find_rule_appns(cls, rule_or_name, ag_shape, sub_ag_shape):
        """Receives:
            rule_or_name 
            ag_shape 
            sub_ag_shape
        Wraps _sgi.find_rule_appns. Returns:
        # Wraps _sgi.find_rule_appns_ag. Returns:
            appns           [appn, ...]
                appn        (t, t_of_a, t_of_b, ag_next)
        """
        pass

    @classmethod
    def is_a_sortal_rule(cls, rule_name):
        """Receives:
            rule_name         str
        If negative, receives a message box from sortal. Returns:
            value           boolean. True if the sortal register contains a 
                            rule with that name. False otherwise (plus 
                            message box from sortal)
        Not tested
        try sortal.rule.type
        """
        value = False
        received_value = _sgi.getRule(rule_name)
        if not received_value == None:
            value = True
        return value

    def delete_empty_layers(self):
        """Deletes all layers that contain no objects, except the default 
        layer. Returns:
            n               int. Number of layers deleted
        """
        layer_names = _rs.LayerNames()
        default = _s.Settings.default_layer_name
        layer_names.remove(default)
        n = 0
        for name in layer_names:
            obs = _rs.ObjectsByLayer(name)
            if obs == []:
                _rs.DeleteLayer(name)
                n += 1
        return n

