from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

import rhinoscriptsyntax as rs
from builtins import range

from sortal import (transformSet, rule, primitiveSort, point3D, point2D, coordinate, lineSegment3D, lineSegment2D, 
                    line3D, label, rWeight, disjunctiveSort, attributeSort, affineTransform,  vector3D)

from mpmath import pi, sin, cos

"""
ConvertS CLASS
Takes sortal metaform as input and converts it into an agnostic form

INPUT
Any metaform, discrete form or interval form; these are sortal objects

RETURNS
Agnostic dictionary form of sortal object with format: {'points': [<p1>, <p2>], 'line segments': [<ls1>, <ls2>]}
""" 

class convertS:
    
    def __init__(self):
        pass

    # Draws form as Rhino objects
    def toAgnostic(self, form):
        self.form = form
        self.type = form.type
        self.objects = {}
        
        # Entered if form is a metaform, i.e. contains both points and line segments
        # Structure of metaform: <metaform name>.elements contains one form of line segments, one form of points
        # To access each individual point or line segment -> since the list stored in the metaform 
        # is constructed as [<point form>, <line segment form>]
        # i.e. <metaform>.name.elements[0].elements -> returns list of points
        # <metaform>.name.elements[1].elements -> returns list of line segments

        if self.type == 'form-metaForm':
            elements = self.form.elements

            for ind in range(len(elements)):
                individuals =  elements[ind].elements
                obj = []
                
                # Converts points
                if elements[ind].type == 'form-discrete':
                    for indiv in individuals:
                        obj.append(self.Point(indiv))     # Extracts data from points
                    type = 'points'
                        
                # Converts lines
                elif elements[ind].type == 'form-interval':
                    obj = self.Line(individuals)          # Extracts data from line segments
                    type = 'line segments'   

                self.objects[type] = obj
        
        # Entered if form only contains points 
        elif self.type == 'form-discrete':
            individuals = self.form.elements[0]
            if type (individuals) == type(self.form.elements):
                
                for indiv in individuals:
                    obj.add(self.Point(indiv))
            else:
                obj.add(self.Point(individuals))
                
            self.objects['points'] = obj
            type = 'points'
        
        # Entered if form only contains line segments
        elif self.type == 'form-interval':
            individuals = self.form.elements
            obj = self.Line(individuals)
            self.objects['line segments'] = obj

        return self.objects
    
    # Extracts data from line segments, returns agnostic tuple ((coordinates), {attributes dictionary}) of line segment data
    def Line(self, line_set):
        v1, v2 = 0, 0
        lines = []
        
        # Extracts attributes and coordinates
        for i in range(len(line_set)):
            v1 = (float(line_set[i].start.x.value), float(line_set[i].start.y.value), float(line_set[i].start.z.value),)
            v2 = (float(line_set[i].end.x.value), float(line_set[i].end.y.value), float(line_set[i].end.z.value),)
            coords = (v1, v2)
            att = {}
            
            # Extracts attributes from line segments inside line_set if there are attributes inside
            # Appends them to text and weight lists with index number corresponding to which line they belong to
            # form-discrete - labels, descriptions
            # form-ordinal - weights
            # form-special - special variant of weights specifically for Rhino, 
            # does operations arithmetically (color, weight)
            
            if line_set[i].hasAttribute():
                ind  =  line_set[i].attribute
                if ind.type == 'form-discrete':
                    if ind.elements[0].getValue[0] == '`':
                        att['descriptions'] = [ind.elements[0].getValue()]
                    else:
                        att['labels']= [ind.elements[0].getValue()]
                    
                elif ind.type == 'form-ordinal' or ind.type == 'form-special':
                    if type(ind.individual.getValue()) == type(1.2) or type(ind.individual.getValue()) == type(1):
                        att['weight'] = [ind.individual.getValue()]
                        
                    elif type(ind.individual.getValue()) == type([]):
                        colorRGB = [ind.individual.getValue()[0], ind.individual.getValue()[1], ind.individual.getValue()[2]]
                        att['color'] = colorRGB
                        
                    elif type(ind.individual.getValue()) == type(1):
                        att['enums'] = [ind.individual.getValue()]
                    
                elif ind.type == 'form-metaForm':

                    for attribute in ind.elements:
                        if attribute.type == 'form-discrete':
                            if attribute.elements[0].getValue()[0] == '`':
                                att['descriptions'] = [attribute.elements[0].getValue()]
                              
                            else:
                                att['labels'] = [attribute.elements[0].getValue()]
                               
                        if attribute.type == 'form-ordinal' or attribute.type == 'form-special':
                            if type(attribute.individual.getValue()) == type(1.2) or type(attribute.individual.getValue()) == type(1):
                                att['weight']= attribute.individual.getValue()
                                
                            elif type(attribute.individual.getValue()) == type([]):
                                att['color'] = [attribute.individual.getValue()[0], attribute.individual.getValue()[1], attribute.individual.getValue()[2]]

                            elif type(attribute.individual.getValue()) == type('a'):
                                att['enums'] = [attribute.individual.getValue()]
            
           
            lines.append((coords, att))

        return lines
    
     
    # Extracts data from points, returns agnostic tuple ((coordinates), {attributes dictionary}) of point data
    def Point(self, pointSource):
        pt = (float(pointSource.value.x.value), float(pointSource.value.y.value), float(pointSource.value.z.value))
        att = {}
        
        # Handles point if it has attributes
        if pointSource.hasAttribute():
            ind  =  pointSource.attribute # Assigns pointSource's attribute
            
            
            # Extracts label or description
            if ind.type == 'form-discrete':
                if ind.elements[0].getValue()[0] == '`':
                    att['descriptions'] = [ind.elements[0].getValue()]
                              
                else:
                   att['labels'] = [ind.elements[0].getValue()]

            # Extracts color of point/text dot OR enumerative value OR weight255
            elif ind.type == 'form-ordinal' or ind.type == 'form-special':
                if type(ind.individual.getValue()) == type(1.0) or type(ind.individual.getValue()) == type(100):
                    att['weight'] = ind.individual.getValue()
                    
                elif type(ind.individual.getValue()) == type(100):
                    att['weight255'] = [ind.individual.getValue(), ind.individual.getValue(), ind.individual.getValue()]
                    
                elif type(ind.individual.getValue()) == type([]):
                    att['color'] = [ind.individual.getValue()[0], ind.individual.getValue()[1], ind.individual.getValue()[2]]

                elif type(ind.individual.getValue()) == type('a'):
                    att['enums'] = ind.individual.getValue()

            # Handles object if it has several attributes
            elif ind.type == 'form-metaForm':
                
                for attribute in ind.elements:
                    
                    if attribute.type == 'form-discrete':
                        if attribute.elements[0].getValue()[0] == '`':
                            att['descriptions'] = [attribute.elements[0].getValue()]
                          
                        else:
                            att['labels'] = [attribute.elements[0].getValue()]
                               
                    if attribute.type == 'form-ordinal' or attribute.type == 'form-special':
                        if type(attribute.individual.getValue()) == type(100) or type(attribute.individual.getValue()) == type(1.0):
                            att['weight255'] = [attribute.individual.getValue(), attribute.individual.getValue(), attribute.individual.getValue()]

                        elif type(attribute.individual.getValue()) == type([]):
                            att['color'] = [attribute.individual.getValue()[0], attribute.individual.getValue()[1], attribute.individual.getValue()[2]]

                        elif type(attribute.individual.getValue()) == type('a'):
                            att['descriptions'] = [attribute.individual.getValue()]
            
                            
        pt = ((pt), att)
            
        return pt