from view import settings as                    _s

class Agnostic(object):
    """A utility for constructing agnostic specs
    """

    def __init__(self):
        pass

    @classmethod
    def new_ag_rule(cls, ag_left, ag_right, name, descr):
        """Receives:
            ag_left         ag_shape
            ag_right        ag_shape
                ag_shape    {'points': points, 'line segments': lines}
            name            str
            descr           str
        Returns:
            ag_rule         ((ag_left, ag_right), attrib_dict)
                attrib_dict {   'name': str, 
                                'description': str}
        """
        ag_rule = (
            (ag_left, ag_right), 
            {   'name': name, 
                'description': descr})
        return ag_rule

    @classmethod
    def new_ag_shapes(cls, shape_specs):
        """Receives:
            shape_specs     [shape_spec, ...]
                shape_spec  (point_specs, line_specs)
                point_spec  (x, y, z, point_labels)
                line_spec   (xyz1, xyz2, line_labels, line_weight)
        Returns:
            ag_shapes       [ag_shape, ...]
                ag_shape    {   'points': ordered_ag_points, 
                                'line segments': ordered_ag_lines}
                ordered_ag_points
                            [ag_point, ...]
                ordered_ag_lines
                            [ag_line, ...]
                ag_line     ((xyz1, xyz2), attrib_dict)
                ag_point    (xyz, attrib_dict)
        """
        ag_shapes = []
        for shape_spec in shape_specs:
            point_specs, line_specs = shape_spec
            ag_shape = Agnostic.new_ag_shape(point_specs, line_specs)
            ag_shapes.append(ag_shape)
        return ag_shapes

    @classmethod
    def new_ag_shape(cls, point_specs, line_specs):
        """Receives:
            point_specs     [point_spec, ...]
            line_specs      [line_spec, ...]
                point_spec  (x, y, z, labels)
                line_spec   (xyz1, xyz2, labels, weight)
        Returns:
            ag_shape        {   'points': sorted_ag_points, 
                                'line segments': sorted_ag_lines}
        """
        ag_points = []
        for spec in point_specs:
            x, y, z, labels = spec
            ag_point = Agnostic.new_ag_point(x, y, z, labels)
            ag_points.append(ag_point)
        ag_lines = []
        for spec in line_specs:
            xyz1, xyz2, labels, weight = spec
            ag_line = Agnostic.new_ag_line(xyz1, xyz2, labels, weight)
            ag_lines.append(ag_line)
        ag_shape = {
            'points': sorted(ag_points), 
            'line segments': sorted(ag_lines), 
        }
        return ag_shape

    @classmethod
    def new_ag_lines(cls, arg_tuple_lists):
        """Receives:
            arg_tuple_lists [arg_list, ...]
                args        [xyz1, xyz2, labels, weight]
                xyz1        (num, num, num)
                xyz2        (num, num, num)
                labels      str or [str, ...]
                weight      num
        Returns:
            ag_lines        [ag_line, ...]
        """
        ag_lines = []
        for args in arg_tuple_lists:
            xyz1, xyz2, labels, weight = args
            ag_line = Agnostic.new_ag_line(xyz1, xyz2, labels, weight)
            ag_lines.append(ag_line)
        return ag_lines
        # return sorted(ag_lines)
    
    @classmethod
    def new_ag_line(
        cls, 
        xyz1, 
        xyz2, 
        labels_in=[], 
        weight=_s.Settings.default_line_weight
    ):
        """Receives:
            xyz1            (num, num, num)
            xyz2            (num, num, num)
            labels_in       str or [str, ...]
            weight          num
        Returns:
            ag_line         (xyz_pair, attrib_dict)
        """
        xyz_pair = (xyz1, xyz2)
        attrib_dict = {
            'labels': labels_in, 
            'weight': weight, 
        }
        if type(labels_in) == str:
            label = labels_in
            labels = [label]
        else:
            labels = labels_in
        attrib_dict['labels'] = labels
        ag_line = (xyz_pair, attrib_dict)
        return ag_line

    @classmethod
    def new_ag_points(cls, arg_tuples):
        """Receives:
            arg_tuples      [arg_tuple, ...]
                arg_tuple   (num, num, num) or 
                            (num, num, num, str) or 
                            (num, num, num, [str, ...])
        Returns:
            ag_points       [ag_point, ...]
        """
        ag_points = []
        for t in arg_tuples:
            x, y, z = t[0], t[1], t[2]
            if len(t) > 3:
                labels = t[3]
            else:
                labels = []
            ag_point = cls.new_ag_point(x, y, z, labels)
            ag_points.append(ag_point)
        return sorted(ag_points)

    @classmethod
    def new_ag_points_from_xyzs(cls, arg_pairs):
        """Receives:
            arg_pairs       (xyz, labels)
                xyz         (num, num, num)
                labels      str or [str, ...]
        Returns:
            ag_points       [ag_point, ...]
                ag_point    (xyz, attrib_dict)
        """
        ag_points = []
        for pair in arg_pairs:
            xyz, labels = pair
            x, y, z = xyz
            ag_point = cls.new_ag_point(x, y, z, labels)
            ag_points.append(ag_point)
        return ag_points

    @classmethod
    def new_ag_point(cls, x, y, z, labels_in=[]):
        """Receives:
            x, y, z         num
            labels_in       str or [str, ...]
        Returns:
            ag_point        (xyz, attrib_dict)
                xyz         (num, num, num)
                attrib_dict {'labels': [str, ...]}
        """
        xyz = (x, y, z)
        attrib_dict = {
            'labels': [], 
        }
        if type(labels_in) == str:
            label = labels_in
            labels = [label]
            attrib_dict['labels'] = labels
        else:
            labels = labels_in
            attrib_dict['labels'] = labels
        ag_point = (xyz, attrib_dict)
        return ag_point

    @classmethod
    def is_empty_ag_shape(cls, item):
        """Receives:
            item            
        Returns:
            value           boolean. True if item is an empty ag_shape. False 
                            otherwise
        """
        value = False
        if (item == {} or
            item == {
                'points': []} or
            item == {
                'line segments': []} or
            item == {
                'points': [], 
                'line segments': []}
        ):
            value = True
        return value


