import                                  random
from view import element as             _e
from view import frame as               _fr
from view import line as                _l
import rhinoscriptsyntax as             _rs
from view import settings as            _s
from view import text_dot as            _td
from view import view_main as           _vm

origin = _s.Settings.origin
red_green = _s.Settings.red_green

class Draw(object):
    def __init__(self):
        pass

    @classmethod
    def draw_framed_object_groups(
        cls, n, overlap=False, ref_point=origin, rel_locations='i', 
    ):
        """Receives:
            n               int. The number of groups, n >= 0
            overlap         boolean. True if consecutive groups overlap; 
                            False otherwise
            ref_point       point3d. The reference point of the first 
                            (leftmost) group
            rel_locations   str. Locations of objects relative to the frame. 
                            A concatenation of:
                f           frame
                i           objects inside the frame
                s           objects straddling the frame
                o           objects outside the frame
        Draws n groups of objects. Each group may have a frame. Each object 
        in a group is inside, straddling, or outside the frame. Returns:
            framed_ob_lists [framed_ob_list, ...]
                framed_ob_list  
                            [guid, ...]. A list of guids of objects and the 
                            frame containing them, if there is one. 
                            Otherwise, a list of all objects. 
        """
        framed_ob_lists = []
        offset = cls._find_group_offset(overlap)
        for i in range(n):
            total_offset = _rs.PointScale(offset, i)
            ref_point_i = _rs.PointAdd(ref_point, total_offset)
            framed_ob_list = (
                cls._draw_framed_object_group(ref_point_i, rel_locations))
            framed_ob_lists.append(framed_ob_list)
        return framed_ob_lists

    @classmethod
    def _find_group_offset(cls, overlap):
        """Receives:
            overlap         boolean. Whether consecutive groups overlap.
        Returns:
            group_offset    point3d. The offset between group reference 
                            points. Less than the frame size, if they 
                            overlap. Greater than the frame size, if they do 
                            not overlap
        """
        ref_point = origin
        if overlap:
            rel_location = 'inside'
        else:
            rel_location = 'outside'
        group_offset = cls._make_random_xyz(ref_point, rel_location)
        return group_offset

    @classmethod
    def draw_framed_element_groups_custom(cls, rel_locations_lists):
        """Receives:
            rel_locations_lists
                            [rel_locations, ...]. A list of relative 
                            locations, where rel_locations is a concatenation 
                            of:
                f           frame
                i           elements inside the frame
                o           elements outside the frame
        Draws framed object groups. Returns:
            els             [guid, ...]. A list of element guids
            frames          [guid, ...]. A list of frame instance guids
        """
        ob_group = []
        unit_offset = [50, 0, 0]
        n = len(rel_locations_lists)
        for i in range(n):
            total_offset = _rs.PointScale(unit_offset, i)
            ref_point_i = _rs.PointAdd(origin, total_offset)
            ob_group_i = (cls._draw_framed_object_group(
                ref_point_i, rel_locations_lists[i]))
            ob_group.extend(ob_group_i)
        els, frames = [], []
        for ob in ob_group:
            if _e.Element.is_element(ob):
                el = ob
                els.append(el)
            elif _fr.Frame.is_an_instance(ob):
                frame = ob
                frames.append(frame)
            else:
                pass
        return (els, frames)

    @classmethod
    def _draw_framed_object_group(cls, ref_point=origin, rel_locations='i'):
        """Receives:
            ref_point       point3d. The reference point
            rel_locations   str. A concatenation of:
                f           frame
                i           objects inside the frame
                s           objects straddling the frame
                o           objects outside the frame
        Draws a group of objects and, optionally, a frame. Returns:
            framed_ob_group [guid, ...]. A list of guids of the frame and of 
                            the objects in the frame, if there is a frame. 
                            Otherwise, a list of guids of all objects. 
        """
        n = 1
        include_all_elements = True
        framed_ob_group = []
        if 'f' in rel_locations:
            current_layer = _rs.CurrentLayer()
            scale = [32, 32, 1]
            frame = _fr.Frame.new_instance(current_layer, ref_point, scale)
            framed_ob_group.append(frame)
            include_all_elements = False
        if 'i' in rel_locations:
            elements, non_elements = cls.draw_objects(
                [   'lines', 
                    'text_dots', 
                    'others', 
                ], 
                n, 
                ref_point, 
                'inside')
            framed_ob_group.extend(elements)
            framed_ob_group.extend(non_elements)
        if 's' in rel_locations:
            elements, non_elements = cls.draw_objects(
                [   'lines', 
                    'text_dots', 
                    'others', 
                ], 
                n, 
                ref_point, 
                'straddling')
            text_dots = _vm.ViewMain._extract_text_dots(elements)
            framed_ob_group.extend(text_dots)
            framed_ob_group.extend(non_elements)
            if include_all_elements:
                lines = _vm.ViewMain._extract_lines(elements)
                framed_ob_group.extend(lines)
        if 'o' in rel_locations:
            elements, non_elements = cls.draw_objects(
                [   'lines', 
                    'text_dots', 
                    'others', 
                ], 
                n, 
                ref_point, 
                'outside')
            if include_all_elements:
                framed_ob_group.extend(elements)
                framed_ob_group.extend(non_elements)
        return framed_ob_group

    @classmethod
    def draw_objects(
        cls, object_types, n=2, ref_point=origin, rel_location='i'
    ):
        """Receives:
            object_types    [<type>, ...]
                <type>      str. 'lines' | 'text_dots' | 'points' | 'curves' | 
                            'polylines' | 'rectangles' | 'others'
            ref_point       point3d
            rel_location    str. Location relative to the frame: 'i' | 
                            'inside' | 's' | 'straddling' | 'o' | 'outside'
        """
        elements, non_elements = [], []
        for type_ in object_types:
            if type_ == 'lines':
                lines = _l.Line.draw_random_lines(
                # lines = _l.Line.draw_random(
                    n, ref_point, rel_location)
                elements.extend(lines)
            elif type_ == 'text_dots':
                text_dots = _td.TextDot.draw_random_text_dots(
                # text_dots = _td.TextDot.draw_random(
                    n, ref_point, rel_location)
                elements.extend(text_dots)
            elif type_ == 'points':
                points = cls.draw_random_points(
                    n, ref_point, rel_location)
                non_elements.extend(points)
            elif type_ == 'curves':
                curves = cls.draw_random_curves(
                    n, ref_point, rel_location)
                non_elements.extend(curves)
            elif type_ == 'polylines':
                polylines = cls.draw_random_polylines(
                    n, ref_point, rel_location)
                non_elements.extend(polylines)
            elif type_ == 'rectangles':
                rectangles = cls.draw_random_rectangles(
                    n, ref_point, rel_location)
                non_elements.extend(rectangles)
            elif type_ == 'others':
                others = cls.draw_random_others(
                    n, ref_point, rel_location)
                non_elements.extend(others)
            else:
                error_message = "Don't recognize the type %s" % type_
                print(error_message)
        return (elements, non_elements)

    @classmethod
    def draw_random_points(cls, n=2, ref_point=origin, rel_location='i'):
        points = []
        for i in range(n):
            point = cls.draw_random_point(ref_point, rel_location)
            points.append(point)
        return points

    @classmethod
    def draw_random_point(cls, ref_point=origin, rel_location='i'):
        xyz = cls._make_random_xyz(ref_point, rel_location)
        point = cls.draw_point(xyz)
        return point

    @classmethod
    def draw_random_others(cls, n=2, ref_point=origin, rel_location='i'):
        others = []
        curves = cls.draw_random_curves(n, ref_point, rel_location)
        polylines = cls.draw_random_polylines(n, ref_point, rel_location)
        lights = cls.draw_random_lights(n, ref_point, rel_location)
        others.extend(curves)
        others.extend(polylines)
        others.extend(lights)
        return others

    @classmethod
    def draw_random_curves(cls, n=2, ref_point=origin, rel_location='i'):
        curves = []
        for i in range(n):
            curve = cls.draw_random_curve(ref_point, rel_location)
            curves.append(curve)
        return curves

    @classmethod
    def draw_random_curve(cls, ref_point=origin, rel_location='i'):
        triple = cls._make_random_xyz_triple(ref_point, rel_location)
        curve = _rs.AddCurve(triple)
        return curve

    @classmethod
    def draw_random_rectangles(cls, n=2, ref_point=origin, rel_location='i'):
        rectangles = []
        for i in range(n):
            rectangle = cls.draw_random_rectangle(ref_point, rel_location)
            rectangles.append(rectangle)
        return rectangles

    @classmethod
    def draw_random_rectangle(cls, ref_point=origin, rel_location='i'):
        x_point = _rs.PointAdd(ref_point, (1, 0, 0))
        y_point = _rs.PointAdd(ref_point, (0, 1, 0))
        plane = _rs.PlaneFromPoints(ref_point, x_point, y_point)
        xyz = cls._make_random_xyz(ref_point, rel_location)
        rectangle = _rs.AddRectangle(plane, xyz[0], xyz[1])
        return rectangle

    @classmethod
    def draw_random_polylines(cls, n=2, ref_point=origin, rel_location='i'):
        polylines = []
        for i in range(n):
            polyline = cls.draw_random_polyline(ref_point, rel_location)
            polylines.append(polyline)
        return polylines

    @classmethod
    def draw_random_polyline(cls, ref_point=origin, rel_location='i'):
        xyzs = cls._make_random_xyz_triple(ref_point, rel_location)
        polyline = _rs.AddPolyline(xyzs)
        return polyline

    @classmethod
    def draw_point(cls, xyz):
        point = _rs.AddPoint(xyz)
        return point

    @classmethod
    def draw_random_lights(cls, n=2, ref_point=origin, rel_location='i'):
        lights = []
        for i in range(n):
            light = cls.draw_random_light(ref_point, rel_location)
            lights.append(light)
        return lights

    @classmethod
    def draw_random_light(cls, ref_point=origin, rel_location='i'):
        xyz = cls._make_random_xyz(ref_point, rel_location)
        light = _rs.AddPointLight(xyz)
        return light

    @classmethod
    def make_random_xyz_pair(
        cls, ref_point=origin, rel_location='i'
    ):
        if rel_location in ['i', 'inside']:
            xyz1 = cls._make_random_xyz(ref_point, 'i')
            xyz2 = cls._make_random_xyz(ref_point, 'i')
        elif rel_location in ['s', 'straddling']:
            xyz1 = cls._make_random_xyz(ref_point, 'i')
            xyz2 = cls._make_random_xyz(ref_point, 'o')
        elif rel_location in ['o', 'outside']:
            xyz1 = cls._make_random_xyz(ref_point, 'o')
            xyz2 = cls._make_random_xyz(ref_point, 'o')
        else:
            error_message = "Should have received 'i', 's', or 'o'"
            print(error_message)
        return (xyz1, xyz2)

    @classmethod
    def draw_origin_marker(cls):
        r = radius = 2
        marker = _rs.AddCircle(origin, r)
        _rs.ObjectColor(marker, red_green)
        _rs.Redraw()

    @classmethod
    def _make_random_xyz_triple(cls, ref_point=origin, rel_location='i'):
        xyz_triple = [
            cls._make_random_xyz(ref_point, rel_location) for i in range(3)]
        return xyz_triple

    @classmethod
    def _make_random_xyz(cls, ref_point=origin, rel_location='i'):
        """Receives:
            ref_point       point3d. The reference point
            rel_location    str. 'i' | 'inside' | s' | 'straddling' | 'o' | 
                            'outside'. The location of the xyz relative to a 
                            frame at the reference point
        Returns:
            xyz             (num, num, num). A triple
        """
        frame_size = _s.Settings.frame_size
        min_ = ref_point
        frame_max = _rs.PointAdd(min_, frame_size)
        max_ = _rs.PointAdd(frame_max, frame_size)
        if rel_location in ['i', 'inside']:
            x = random.uniform(min_[0], frame_max[0])
            y = random.uniform(min_[1], frame_max[1])
            z = random.uniform(min_[2], frame_max[2])
        elif rel_location in ['s', 'straddling']:
            x = random.uniform(min_[0], frame_max[0])
            y = random.uniform(min_[1], frame_max[1])
            z = min_[2]
        elif rel_location in ['o', 'outside']:
            x = random.uniform(frame_max[0], max_[0])
            y = random.uniform(frame_max[1], max_[1])
            z = random.uniform(frame_max[2], max_[2])
        else:
            error_message = "Should have received 'i', 's', or 'o'"
            print(error_message)
        return (x, y, z)

    @classmethod
    def clear_all(cls):
        _vm.ViewMain.clear_all()





























