from view import frame as               _fr
from view import line as                _l
import rhinoscriptsyntax as             _rs
from view import settings as            _s
from view import text_dot as            _td

class Element(object):
    """Rhino object: line or text dot. Called 'Element' to avoid 'Object'
    """

    def __init__(self):
        pass

    @classmethod
    def is_element(cls, object_):
        """Receives:
            object_         guid
        Returns:
            value           boolean. True if object_ is a line or a text dot. 
                            False otherwise
        """
        value = False
        if (_l.Line.is_line (object_) or
            _td.TextDot.is_text_dot(object_)
        ):
            value = True
        return value

    @classmethod
    def extract_text_dots_and_lines(cls, obs):
        """Receives:
            obs             [guid, ...]. A list of object guids
        Returns:
            text_dots       [guid, ...]. A list of text dot guids
            lines           [guid, ...]. A list of line guids
        """
        text_dots = []
        lines = []
        for ob in obs:
            if _td.TextDot.is_text_dot(ob):
                text_dot = ob
                text_dots.append(text_dot)
            elif _l.Line.is_line(ob):
                line = ob
                lines.append(line)
        return (text_dots, lines)

    @classmethod
    def extract_from_objects(cls, obs):
        """Receives:
            obs             [guid, ...]. A list of object guids
        Returns:
            els             [guid, ...]. A list of element guids. May be empty
        """
        els = []
        for ob in obs:
            if cls.is_element(ob):
                el = ob
                els.append(el)
        return els

    @classmethod
    def extract_with_frames_from_objects(cls, obs):
        """Receives:
            obs             [guid, ...]. A list of object guids. Must not be 
                            empty. Guaranteed by calling function
        Returns:
            els             [guid, ...]. A list of element guids. May be 
                            empty
            frames          [guid, ...]. A list of frame instance guids. May 
                            be empty
        """
        els = []
        frames = []
        for ob in obs:
            if cls.is_element(ob):
                els.append(ob)
            elif _fr.Frame.is_an_instance(ob):
                frames.append(ob)
            else:
                pass
        return (els, frames)

    @classmethod
    def extract_by_frame(cls, elements, frames):
        """Receives:
            elements        [guid, ...]. A list of element guids
            frames          [guid, guid]. A list of 2 non-overlapping frame 
                            instance guids
        Returns:
            els_a           [guid, ...]. A list of element guids in frame a
            els_b           [guid, ...]. A list of element guids in frame b
            None            if unsuccessful
        """
        frame_a, frame_b = frames
        els_a, els_b = [], []
        for el in elements:
            if _fr.Frame.contains_object(frame_a, el):
                els_a.append(el)
            elif _fr.Frame.contains_object(frame_b, el):
                els_b.append(el)
        return (els_a, els_b)

    @classmethod
    def find_elements(cls):
        """Returns:
            elements        [guid, ...]. A list of guids of elements (lines, 
                            text dots) in the grammar
        """
        filter_ = (
            _s.Settings.curve_filter +
            _s.Settings.text_dot_filter
        )
        curves_and_text_dots = _rs.ObjectsByType(filter_)
        elements = []
        for obj in curves_and_text_dots:
            if cls.is_element(obj):
                element = obj
                elements.append(element)
        return elements

