import rhinoscriptsyntax as             _rs
from view import settings as            _s
from view import view_utilities as      _vu

class Frame(object):
    def __init__(self):
        pass

    @classmethod
    def new_instance(cls, layer_name, position, scale=(1, 1, 1)):
        """Receives:
            layer_name      str. The name of an existing layer
            position        Point3d. The position of the frame block
            scale           [num, num, num]. The scale factors
        Creates a frame block definition, if there is not already one. Inserts 
        a frame block instance at the specified position, at the specified 
        scale, and on the specified layer. Returns:
            frame_instance  guid. The guid of the new frame block instance
        """
        frame_name = _s.Settings.frame_name
        default_layer_name = _s.Settings.default_layer_name
        if not cls._definition_exists():
            cls._new_definition()
        _rs.CurrentLayer(layer_name)
        frame_instance = _rs.InsertBlock(frame_name, position, scale)
        _rs.CurrentLayer(default_layer_name)
        return frame_instance

    @classmethod
    def _definition_exists(cls):
        """Returns:
            value           boolean. True, if a frame block definition exists. 
                            False, otherwise
        """
        block_name = _s.Settings.frame_name
        value = _rs.IsBlock(block_name)
        return value

    @classmethod
    def _new_definition(cls):
        """Precondition:
            A frame definition does not already exist
            A layer named <frame_layer_name> does not already exist
        Creates a frame definition on its own layer. Returns:
            frame_name_out  str. The name of the frame definition
        """
        layer_name = _s.Settings.frame_layer_name
        color_name = _s.Settings.frame_color_name
        _rs.AddLayer(layer_name, color_name)
        _rs.CurrentLayer(layer_name)

        line_guids = cls._get_guids_3d()
        base_point = _s.Settings.frame_base_point
        frame_name_in = _s.Settings.frame_name
        delete_input = True
        frame_name_out = _rs.AddBlock(
            line_guids, base_point, frame_name_in, delete_input)

        default_layer_name = _s.Settings.default_layer_name
        _rs.CurrentLayer(default_layer_name)
        return frame_name_out

    @classmethod
    def _get_guids_3d(cls):
        """Draws a 3D shape frame. Returns:
            line_guids      [guid, ...]. A list of the guids of the lines in 
                            the frame
        """
        point_pairs = [
            ((0, 0, 0), (0, 0, 1)), 
            ((0, 0, 0), (0, 1, 0)), 
            ((0, 0, 0), (1, 0, 0)), 
            ((0, 0, 1), (0, 1, 1)), 
            ((0, 0, 1), (1, 0, 1)), 
            ((0, 1, 0), (0, 1, 1)), 
            ((0, 1, 0), (1, 1, 0)), 
            ((0, 1, 1), (1, 1, 1)), 
            ((1, 0, 0), (1, 1, 0)), 
            ((1, 0, 0), (1, 0, 1)), 
            ((1, 0, 1), (1, 1, 1)), 
            ((1, 1, 0), (1, 1, 1)), 
        ]
        line_guids = []
        for pair in point_pairs:
            guid = _rs.AddLine(pair[0], pair[1])
            line_guids.append(guid)
        return line_guids

    @classmethod
    def _get_guids_2d(cls):
        """Draws a 2D shape frame. Returns:
            line_guids      [guid, ...]. A list of the guids of the lines in 
                            the frame
        """
        p00, p01, p10, p11 = (0, 0, 0), (0, 1, 0), (1, 0, 0), (1, 1, 0)
        point_pairs = [
            (p00, p01), 
            (p00, p10), 
            (p01, p11), 
            (p10, p11), 
        ]
        line_guids = []
        for pair in point_pairs:
            guid = _rs.AddLine(pair[0], pair[1])
            line_guids.append(guid)
        return line_guids

    @classmethod
    def find_padded_frame_specs(cls, bounding_points):
        """Receives:
            bounding_points [point3d, ...]. A list of the 8 vertices of the 
                            bounding box. The list is empty for the empty 
                            shape
        Finds the specs of a padded frame instance based on the bounding box. 
        If bounding_points is empty (i.e., bounds the empty shape), finds the 
        specs of a minimal frame about the origin. Returns:
            insert_point    point3d
            scale           (num, num, num)
        """
        def _find_vector_length(p0, p1):
            vector = _rs.VectorCreate(p1, p0)
            vector_length = _rs.VectorLength(vector)
            return vector_length

        bpp = bounding_points
        if bpp == None:
            return_value = None
        elif bpp == []:
            insert_point = _vu.get_point3d((-10, -10, 0))
            scale = (20, 20, 1)
            return_value = (insert_point, scale)
        else:
            box_dx = _find_vector_length(bpp[0], bpp[1])
            box_dy = _find_vector_length(bpp[0], bpp[3])
            box_dz = _find_vector_length(bpp[0], bpp[4])
            pad_k = _s.Settings.frame_padding_factor
            pad_min = _s.Settings.frame_padding

            pad_x = box_dx * pad_k
            if pad_x < pad_min:
                pad_x = pad_min
            pad_y = box_dy * pad_k
            if pad_y < pad_min:
                pad_y = pad_min
            pad_z = box_dz * pad_k
            if 0 < pad_z < pad_min:
                pad_z = pad_min
            insert_point = _rs.PointSubtract(bpp[0], [pad_x, pad_y, 0])

            frame_dx = box_dx + (pad_x * 2)
            frame_dy = box_dy + (pad_y * 2)
            frame_dz = box_dz + pad_z
            if frame_dz == 0:
                frame_dz = 1
            scale = (frame_dx, frame_dy, frame_dz)
            return_value = (insert_point, scale)
        return return_value

    @classmethod
    def find_pad(cls, box_length):
        """Finds the pad for one dimension (x, y, or z) of a bounding box.
        Receives:
            box_length      num. num >= 0. In one dimension (x, y, or z)
        Returns:
            pad             num. num >= 0. None if unsuccesful
        """
        if box_length < 0:
            return_value = None
        else:
            pad_k = _s.Settings.frame_padding_factor
            pad_min = _s.Settings.frame_pad_minimum
            pad = box_length * pad_k
            if pad < pad_min:
                pad = pad_min
            return_value = pad
        return return_value

    @classmethod
    def extract_instances(cls, obs):
        """Receives:
            obs             [guid, ...]
        Returns:
            instrances      [guid, ...]. The guids of 2d frames, i.e., 
                            polylines
        """
        instances = []
        for ob in obs:
            if cls.is_an_instance(ob):
                instances.append(ob)
        return instances

    @classmethod
    def is_an_instance(cls, item):
        """Receives:
            item            guid
        Returns:
            value           boolean. True if the item is a Frame instance. 
                            False otherwise
        """
        value = False
        if (_rs.ObjectType(item) == _s.Settings.block_instance_filter and
            _rs.BlockInstanceName(item) == _s.Settings.frame_name
        ):
            value = True
        return value

    @classmethod
    def find_instance_position(cls, guid):
        """Receives:
            guid            guid. The guid of the frame instance
        Returns:
            position        Point3d. The position of the frame instance
        """
        position = _rs.BlockInstanceInsertPoint(guid)
        return position

    @classmethod
    def get_instance_position_by_name(cls, name):
        """Receives:
            name            str. The name of a frame instance on the current 
                            layer
        Returns:
            position        Point3d. The insertion point of the frame instance
        """
        guid = cls.get_instance_by_name(name)
        position = cls.find_instance_position(guid)
        return position

    @classmethod
    def get_frame_position_from_user(cls):
        """Returns:
            origin          Point3d. The position of the frame instance. Must 
                            be in the xy plane
        """
        message_1 = "Select a point in the xy plane"
        message_2 = "The point must be in the xy plane. Try again"
        origin = _rs.GetPoint(message_1)
        while not origin[2] == 0:
            origin = _rs.GetPoint(message_2)
        return origin

    @classmethod
    def get_bounding_box_by_name(cls, name):
        """Receives:
            name            str. The name of a frame instance on the current 
                            layer
        Returns:
            box             [p1, ..., p8], if successful. None otherwise
                pi          Point3d
        """
        instance = cls.get_instance_by_name(name)
        if not instance:
            value = None
        else:
            box = _rs.BoundingBox(instance)
            value = box
        return value

    @classmethod
    def get_instance_by_name(cls, name):
        """Receives:
            name            str. The name of a frame instance on the current 
                            layer
        Returns:
            instance        guid. Of the named frame instance, if successful. 
                            None otherwise
        """
        if not type(name) == str:
            value = None
        else:
            named_objects = _rs.ObjectsByName(name)
            instances = []
            for object_i in named_objects:
                if _rs.ObjectLayer(object_i) == _rs.CurrentLayer():
                    instances.append(object_i)
            n_instances = len(instances)
            if n_instances == 0:
                print('Found no frame instances named %s' % name)
                value = None
            elif n_instances == 1:
                instance = instances.pop()
                value = instance
            else:
                instance = _rs.GetObject(
                    'Select the instance you want', 
                    _s.Settings.block_instance_filter)
                value = instance
        return value

    @classmethod
    def extract_from_elements(cls, elems):
        """Receives:
            elems           [guid, ...]. A list of element guids
        Returns:
            ordered_frames  [guid, ...]. A list of frame guids, ordered by 
                            increasing insertion point
        """
        ordered_frames = []
        i_pts = insertion_points = []
        i_pt_frame_dict = {}
        for elem in elems:
            if cls.is_an_instance(elem):
                frame = elem
                i_pt = _rs.BlockInstanceInsertPoint(frame)
                i_pt_frame_dict[i_pt] = frame
                i_pts.append(i_pt)
        i_pts = i_pt_frame_dict.keys()
        sorted_i_pts = sorted(i_pts)
        for p in sorted_i_pts:
            frame = i_pt_frame_dict[p]
            ordered_frames.append(frame)
        return ordered_frames

    @classmethod
    def order_pair(cls, frames):
        """Receives:
            frames          [guid, guid]. A list of 2 disjoint frame instance 
                            guids
        Returns:
            ordered_frames  [guid, guid]. A list of 2 frame instance guids 
                            ordered by insertion point
            None            if unsuccesful
        """
        frame_a, frame_b = frames
        i_pt_a = cls.find_instance_position(frame_a)
        i_pt_b = cls.find_instance_position(frame_b)
        if i_pt_a < i_pt_b:
            ordered_frames = [frame_a, frame_b]
        elif i_pt_b < i_pt_a:
            ordered_frames = [frame_b, frame_a]
        else:
            ordered_frames = None
        return ordered_frames

    @classmethod
    def contains_object(cls, frame, ob):
        """Receives:
            frame           guid. A frame instance
            ob              guid. An object guid
        Returns:
            value           boolean. True if the object lies inside the 
                            frame. False otherwise.
        """
        value = False
        frame_box = _rs.BoundingBox(frame)
        if _rs.IsObjectInBox(ob, frame_box):
            value = True
        return value

    @classmethod
    def overlaps_object(cls, frame, ob):
        """Receives:
            frame           guid. A frame instance
            ob              guid. An object
        Returns:
            value           boolean. True if the frame and the object overlap 
                            or if the frame contains the object. False 
                            otherwise
        """
        value = False
        frame_box = _rs.BoundingBox(frame)
        test_mode = False
        if _rs.IsObjectInBox(ob, frame_box, test_mode):
            value = True
        return value

