from view import element as             _e
from view import frame as               _fr
from controller import guids_to_dat as  _gd
from view import layer as               _l
from view import labeled_arrow as       _la
import rhinoscriptsyntax as             _rs
from view import settings as            _s

class Grammar(object):
    """A grammar contains shapes and rules. Each shape and rule is on its own 
    layer. The layer is named what? A new shape is drawn on its own layers, 
    automatically named. 
    """
    def __init__(self):
        pass

    ### new
    @classmethod
    def set_up_grammar(cls):
        cls.clear_all()
        cls._set_up_first_initial_shape()
        cls._set_up_first_rule()

    @classmethod
    def _set_up_first_initial_shape(cls):
        """Required state:
            There must be no layer with the first initial shape layer name
        Adds a new layer with the first initial shape layer name and one 
        frame instance. Should be executed only once. Returns:
            layer_name      str. The name of the layer, if successful. None, 
                            otherwise
        """
        method_name = '_set_up_first_initial_shape'
        try:
            layer_name = _s.Settings.first_initial_shape_layer_name
            (   layer_name_is_in_use
            ) = (
                _rs.IsLayer(layer_name))
            if layer_name_is_in_use:
                raise ValueError
        except ValueError:
            message = "The layer name '%s' is in use" % layer_name
            print("%s.%s:\n    %s" % (cls.__name__, method_name, message))
            return_value = None
        else:
            frame_instance_position = (
                _s.Settings.first_initial_shape_frame_position)
            return_value = cls._set_up_initial_shape(
                layer_name, frame_instance_position)
        finally:
            return return_value

    @classmethod
    def set_up_subsequent_initial_shape(cls):
        """Prompts the user for a valid layer name and a valid point. Adds a 
        new layer with that name and one frame instance at that point. 
        Returns:
            layer_name      str. The name of the layer, if successful. None,
                            otherwise
        """
        (   layer_name, 
            frame_position
        ) = (
            _l.Layer.get_layer_name_from_user(), 
            _fr.Frame.get_frame_position_from_user())
        return_value = cls._set_up_initial_shape(layer_name, frame_position)
        return return_value

    @classmethod
    def _set_up_initial_shape(cls, layer_name, frame_position):
        """Receives:
            layer_name      str. A well-formed and available layer name
            frame_position  Point3d. The position of the frame instance
        Creates a new layer with the specified name. Inserts a frame instance 
        on the new layer at the specified position. Returns:
            layer_name_out  str. The name of the layer
        """
        _l.Layer.new(layer_name)
        frame_instance_out = _fr.Frame.new_instance(layer_name, frame_position)
        layer_name_out = _rs.ObjectLayer(frame_instance_out)
        return layer_name_out

    @classmethod
    def _set_up_first_rule(cls):
        """Required state:
            There must be no layer with the first rule layer name
        Adds a new layer with the first rule layer name and two frame 
        instances. Should be executed only once. Returns:
            layer_name      str. The name of the layer, if successful. None, 
                            otherwise
        """
        method_name = '_set_up_first_rule'
        try:
            layer_name = _s.Settings.first_rule_layer_name
            (   layer_name_is_in_use
            ) = (
                _rs.IsLayer(layer_name))
            if layer_name_is_in_use:
                raise ValueError
        except ValueError:
            message = "The layer name '%s' is in use" % layer_name
            print("%s.%s:\n    %s" % (cls.__name__, method_name, message))
            return_value = None
        else:
            (   left_frame_position
            ) = (
                _s.Settings.first_rule_left_frame_position)
            return_value = cls._set_up_rule(layer_name, left_frame_position)
        finally:
            return return_value
        
    @classmethod
    def set_up_subsequent_rule(cls):
        """Prompts the user for a valid layer name and a valid point. Adds a 
        new layer with that name and two frame instances relative to that 
        point. Returns:
            layer_name      str. The name of the layer, if successful. None,
                            otherwise
        """
        (   layer_name,
            frame_position
        ) = (
            _l.Layer.get_layer_name_from_user(),
            _fr.Frame.get_frame_position_from_user())
        return_value = cls._set_up_rule(layer_name, frame_position)
        return return_value

    @classmethod
    def _set_up_rule(cls, layer_name, left_frame_position):
        """Receives:
            layer_name      str. A well-formed and available layer name
            left_frame_position
                            Point3d. The position of the left frame instance
        Creates a new layer with the specified name. Inserts two frame 
        instances and a labeled arrow group on the new layer. Returns:
            layer_name_out  str. The name of the rule layer
        """
        _l.Layer.new(layer_name)
        right_frame_position = _s.Settings.get_right_frame_position(
            left_frame_position)
        arrow_position = _s.Settings.get_arrow_position(left_frame_position)
        left_frame_out = _fr.Frame.new_instance(layer_name, left_frame_position)
        right_frame_out = _fr.Frame.new_instance(
            layer_name, right_frame_position)
        labeled_arrow_out = _la.LabeledArrow.new(layer_name, arrow_position)
        return_value = False
        if (left_frame_out and
            right_frame_out and
            labeled_arrow_out
        ):
            return_value = layer_name
        return return_value

    @classmethod
    def set_up_arithmetic(cls):
        """Required state:
            ?
        Prompt the user for a valid layer name and a location. Does not 
        delete anything. Set up: 
            A new layer
            A frame for A 
            A frame for B 
            A frame for A + B
            A frame for A - B
            A frame for B - A
        Return:
            layer_name      str. The name of the layer, if successful. None 
                            otherwise
        """
        layer_name = _l.Layer.get_layer_name_from_user()
        layer_name_out = _l.Layer.new(layer_name)
        _rs.CurrentLayer(layer_name)
        local_origin = _fr.Frame.get_frame_position_from_user()
        frame_names = ['a', 'b', 'a+b', 'a-b', 'b-a']
        for frame_name in frame_names:
            cls._set_up_arithmetic_frame(frame_name, layer_name, local_origin)
        return layer_name_out

    @classmethod
    def _set_up_arithmetic_frame(cls, frame_name, layer_name, local_origin):
        """Receive:
            frame_name      str. 'a' | 'b' | 'a+b' | 'a-b' | 'b-a'
            layer_name      str. A valid name for the layer on which to set up
            local_origin    Point3d. The local origin of the cluster, i.e., 
                            the position of frame a
        Return:
            frame_instance  guid. The guid of the new frame block instance, 
                            if successful
        """
        tag_texts = {
            'a': _s.Settings.arithmetic_tag_text_a, 
            'b': _s.Settings.arithmetic_tag_text_b, 
            'a+b': _s.Settings.arithmetic_tag_text_sum_ab, 
            'a-b': _s.Settings.arithmetic_tag_text_diff_ab, 
            'b-a': _s.Settings.arithmetic_tag_text_diff_ba}
        tag_text = tag_texts[frame_name]
        (frame_position, tag_position) = (
            _s.Settings.get_arithmetic_frame_position(
                frame_name, local_origin))
        frame_instance = _fr.Frame.new_instance(layer_name, frame_position)
        _rs.ObjectName(frame_instance, frame_name)
        _rs.CurrentLayer(layer_name)
        text_height = _s.Settings.arithmetic_tag_text_height
        _rs.AddText(tag_text, tag_position, text_height, justification=2)
        return frame_instance

    @classmethod
    def current_layer_has_arithmetic_set_up(cls):
        """Returns:
            value           True, if the current layer contains an 
                            arithmetic set-up. False otherwise
        """
        value = True
        tags = _rs.ObjectsByType(_s.Settings.annotation_type)
        tag_texts = [_rs.TextObjectText(tag) for tag in tags]
        tag_text_a = _s.Settings.arithmetic_tag_text_a
        tag_text_b = _s.Settings.arithmetic_tag_text_b
        tag_text_sum_ab = _s.Settings.arithmetic_tag_text_sum_ab
        tag_text_diff_ab = _s.Settings.arithmetic_tag_text_diff_ab
        tag_text_diff_ba = _s.Settings.arithmetic_tag_text_diff_ba
        block_instances = _rs.ObjectsByType(_s.Settings.block_type)
        if not (
            len(block_instances) == 5 and
            cls._blocks_are_all_frames(block_instances)
        ):
            value = False
            print('There should be 5 frames')
        elif not (
            len(tag_texts) == 5 and
            tag_text_a in tag_texts and
            tag_text_b in tag_texts and
            tag_text_sum_ab in tag_texts and
            tag_text_diff_ab in tag_texts and
            tag_text_diff_ba in tag_texts
        ):
            value = False
            print("There should be 5 tags: 'a', 'b', 'a+b', 'a-b', and 'b-a'")
        else:
            pass
        return value

    @classmethod
    def get_elements_by_frame_name(cls, name):
        """Receives:
            name            str. The name of a frame instance on the current 
                            layer
        Returns:
            elements_in_frame
                            [guid, ...]. A list of guids of elements (lines 
                            and text dots, not the frame instance) inside the 
                            frame, if successful. None otherwise
        """
        frame_box = _fr.Frame.get_bounding_box_by_name(name)
        if not frame_box:
            value = None
        else:
            objects_on_layer = _rs.ObjectsByLayer(_rs.CurrentLayer())
            elements_in_frame = []
            for object_i in objects_on_layer:
                if ((   _rs.IsCurve(object_i) or
                        _rs.IsTextDot(object_i)
                    ) and
                    _rs.IsObjectInBox(object_i, frame_box)
                ):
                    elements_in_frame.append(object_i)
            value = elements_in_frame
        return value

    @classmethod
    def _get_input_shapes(cls):
        """Reads the input shapes on the current layer. Returns:
            (shape_a, shape_b)
                            (Shape, Shape)
        """
        shape_a = cls._get_shape_by_frame_name('a')
        shape_b = cls._get_shape_by_frame_name('b')
        return (shape_a, shape_b)

    @classmethod
    def _get_shape_by_frame_name(cls, frame_name):
        """Receives:
            frame_name      str. The name of a frame on the current layer
        Returns:
            lshape          LabeledShape. The labeled shape specified by the 
                            elements in the frame, if successful. None 
                            otherwise
        """
        frame = _fr.Frame.get_instance_by_name(frame_name)
        guids = _gd.GuidsToDat._get_elements(frame)
        line_guids, lpt_guids = cls._sort_guids(guids)
        line_specs = cls._get_line_specs_from_guids(line_guids)
        lpt_specs = cls._get_lpoint_specs_from_guids(lpoint_guids)
        lshape = ls.LabeledShape.from_specs(line_specs, lpt_specs)
        #   new argument: single spec list
        return lshape

    @classmethod
    def _sort_guids(cls, guids):
        """Receives:
            guids           [guid, ...]. A list of guids
        Returns:
            line_guids      [guid, ...]. A list of line guids
            lpoint_guids    [guid, ...]. A list of lpoint guids
        """
        method_name = '_sort_guids'
        try:
            if not type(guids) == list:
                raise TypeError
        except TypeError:
            message = 'The argument must be a list'
            cls._print_error_message(method_name, message)
            value = None
        else:
            line_guids = []
            lpoint_guids = []
            for guid in guids:
                if _rs.IsCurve(guid):
                    line_guids.append(guid)
                elif _rs.IsTextDot(guid):
                    lpoint_guids.append(guid)
                else:
                    pass
            value = (line_guids, lpoint_guids)
        finally:
            return value

    @classmethod
    def _get_line_specs_from_guids(cls, guids, local_origin):
        """Receives:
            guids           [guid, ...]. A list of line guids. Guaranteed
            local_origin    Point3d
        >>> Should it reduce overlapping lines?
        Returns:
            specs           [spec, ...], where:
                spec        ((num, num, num), (num, num, num)). Relative to 
                            the local_origin
        """
        specs = []
        for guid in guids:
            spec = cls._get_line_spec_from_guid(guid, local_origin)
            specs.append(spec)
        return specs

    @classmethod
    def _get_line_spec_from_guid(cls, line_guid, local_origin):
        """Receives:
            line_guid       A line guid. Guaranteed
            local_origin    Point3d
        Returns:
            spec            ((num, num, num), (num, num, num)). Relative to 
                            the local origin, if successful. None otherwise
        """
        if not _rs.IsCurve(line_guid):
            value = None
        else:
            p1, p2 = _rs.CurvePoints(line_guid)
            p1_coords = (p1[0], p1[1], p1[2])
            p2_coords = (p2[0], p2[1], p2[2])
            if p1_coords < p2_coords:
                line_spec = (p1_coords, p2_coords)
            else:
                line_spec = (p2_coords, p1_coords)
            value = line_spec
        return value

    @classmethod
    def _get_lpoint_specs_from_guids(cls, guids):
        """Receives:
            guids           [guid, ...]. All types?
        Returns:
            lpoint_specs    [lpoint_spec, ...], where:
                lpoint_spec (num, num, num, str)
        """
        pass

    @classmethod
    def _calculate_shapes(cls, a, b):
        """Receives:
            a, b            Shapes
        Returns:
            (sum_ab, diff_ab, diff_ba)
                sum_ab      Shape. a + b
                diff_ab     Shape. a - b
                diff_ba     Shape. b - a
            None            If not successful
        """
        sum_ab = a + b
        diff_ab = a - b
        diff_ba = b - a
        return (sum_ab, diff_ab, diff_ba)

    @classmethod
    def _draw_output_shapes(cls, shapes):
        """Receives:
            shapes          [Shape, ...]
        Draws each shape in its frame. Returns:
            value           boolean. True if successful. False otherwise
        """
        draw(sum_ab)
        draw(diff_ab)
        draw(diff_ba)
        pass

    @classmethod
    def _blocks_are_all_frames(cls, block_instances):
        """Receives:
            block_instances [Block, ...]. A list of block instances
        Returns:
            value           True, if each is an instance of a frame block. 
                            False otherwise
        """
        value = True
        for block_instance in block_instances:
            if not (
                _rs.BlockInstanceName(block_instance) == _s.Settings.frame_name
            ):
                value = False
                break
        return value

    ### export
    @classmethod
    def export(cls):                            ##  cf import deriv
        """Writes the grammar's dat string to a file, proposing the Rhino 
        document name as the file name. Returns:
            dat_string      str. The dat string, if successful. None, 
                            otherwise
        """
        initial_shapes, rules = cls._get_element_layers()
        if (initial_shapes == [] or 
            rules == []
        ):
            (   error_message
            ) = (
                ' '.join([
                    "The grammar must have at least", 
                    "one initial shape and one rule"]))
            print(error_message)
            return_value = None
        else:
            (   suggested_file_name, 
                dat_string
            ) = (
                _rs.DocumentName(), 
                _gd.GuidsToDat.get_dat_string(initial_shapes, rules))
            dat_string_out = cls._write_to_file(
                suggested_file_name, dat_string)
            return_value = dat_string_out
        return return_value

    @classmethod
    def _get_element_layers(cls):
        """Returns:
            initial_shapes  [str, ...]. A list of the names of layers 
                            containing one frame instance
            rules           [str, ...]. A list of the names of layers 
                            containing two frame instances
        """
        initial_shapes, rules = [], []
        layer_names = _rs.LayerNames()
        for name in layer_names:
            if _l.Layer.contains_initial_shape(name):
                initial_shapes.append(name)
            elif _l.Layer.contains_rule(name):
                rules.append(name)
        return (initial_shapes, rules)

    @classmethod
    def _write_to_file(cls, file_name_in, dat_string):
        title = 'Save grammar'
        file_filter = 'DAT file (*.dat)|*.dat||'
        folder = None
        file_name = file_name_in
        extension = 'dat'
        file_name_out = (
            _r_s.SaveFileName(title, file_filter, folder, file_name, extension))
        if not file_name_out: 
            return
        file = open(file_name_out, "w" )
        file.write(dat_string)
        file.close()
        print(dat_string)

    ####

    @classmethod
    def get_labeled_shape_names(cls):
        """The grammar is guaranteed to be well-formed. Returns:
            labeled_shape_names
                            [str, ...]. A list of the names of the labeled 
                            shapes in initial shapes or rules, i.e., name, 
                            name_L, or name_R
        """
        labeled_shape_names = []
        layer_names = _rs.LayerNames()
        for layer_name in layer_names:
            if _l.Layer.contains_initial_shape(layer_name):
                labeled_shape_names.append(layer_name)
            elif _l.Layer.contains_rule(layer_name):
                left_name = "%s_L" % layer_name
                right_name = "%s_R" % layer_name
                labeled_shape_names.extend([left_name, right_name])
        return labeled_shape_names

    @classmethod
    def remove_from_initial_shapes(cls):
        pass

    @classmethod
    def remove_from_rule(cls):
        pass

    ### import 
    # @classmethod
    # def import(cls):                            ##  can't use this word
        # pass

    # @classmethod
    # def import_derivation(cls):
        # """Prompts the user for a drv file. Draws the derivation
        # Put this in dats_to_guid?
        # """
        # draw_derivation(shape_specs, rules)

    ### new for Rudi
    # @classmethod # pending
    # def contains_one_well_formed_rule(cls):
        # """Returns:
        #     value           boolean. True if the grammar contains at least 
        #                     one well-formed rule; i.e., a layer with a left 
        #                     non-empty framed form and a (possibly empty) 
        #                     right framed form. False otherwise
        # """
        # value = False
        # for layer in layers:
        #     if _l.Layer.contains_well_formed_rule(layer):
        #         value = True
        #         break
        # return value

    @classmethod
    def is_empty(cls):
        """Returns:
            value           boolean. True if there are no elements (lines, 
                            text dots). False otherwise
        """
        value = False
        elements = _e.Element.find_elements()
        if len(elements) == 0:
            value = True
        return value

    ### utilities
    @classmethod                                ##  called
    def clear_all(cls):                         ##  system-created blocks 
                                                ##  only?
        cls._clear_objects()
        cls._clear_blocks()
        cls._clear_layers()
        cls._clear_groups()
        cls._clear_data()

    @classmethod
    def _clear_objects(cls):
        """Deletes all drawn objects, including locked objects, if any. 
        Returns:
            n_objects       int. The number of objects deleted, if successful
        """
        dummy_point = _rs.AddPoint(0, 0, 0)
        value = _rs.DeleteObject(dummy_point)
        include_lights = True
        include_grips = True
        objects = _rs.AllObjects(include_lights, include_grips)
        n_objects = _rs.DeleteObjects(objects)
        remaining_objects = _rs.AllObjects(include_lights, include_grips)
        _rs.UnlockObjects(remaining_objects)
        m_objects = _rs.DeleteObjects(remaining_objects)
        return n_objects + m_objects

    @classmethod                                ##  called
    def _clear_blocks(cls):
        block_names = _rs.BlockNames()
        for name in block_names:
            _rs.DeleteBlock(name)

    @classmethod                                ##  called
    def _clear_layers(cls):                     ##  clear only user layers
        """Deletes all layers. Leaves Default layer
        """
        default_layer_name = _s.Settings.default_layer_name
        layer_names = _rs.LayerNames()
        if not default_layer_name in layer_names:
            _rs.AddLayer(default_layer_name)
        _rs.CurrentLayer(default_layer_name)
        for layer_name in layer_names:
            if not layer_name == default_layer_name:
                value = _rs.DeleteLayer(layer_name)
                # print('Cleared layer %s: %s' % (layer_name, value))

    @classmethod
    def _clear_groups(cls):
        """Deletes all groups 
        """
        groups = _rs.GroupNames()
        if groups:
            for group in groups:
                _rs.DeleteGroup(group)

    @classmethod                                ##  called
    def _clear_data(cls):
        _rs.DeleteDocumentData()

    @classmethod
    def _print_error_message(cls, method_name, message):
        print '%s.%s:\n    %s' % (cls.__name__, method_name, message)

