from view import frame as                       _f
import rhinoscriptsyntax as                     _rs
from view import settings as                    _s

class Layer(object):
    """The layer name is the rule id. So the layer name register is the rule 
    id register
    """
    prohibited_characters = [' ', '#']

    def __init__(self):
        pass

    @classmethod                                ##  done 08-05
    def new(cls, name_in):
        """Receives:
            name_in         str. A well-formed and available layer name
        Adds a layer named <name_in>. Returns:
            name_out        str. The name of the layer
        """
        name_out = _rs.AddLayer(name_in)
        return name_out

    @classmethod
    def make_default(cls):
        """Creates a new default layer (i.e., with an English name) if it does 
        not already exist. Returns:
            layer_name      str. The name of the default layer. None if 
                            unsuccessful
        """
        return_value = None
        default_layer_name = _s.Settings.default_layer_name
        layer_names = _rs.LayerNames()
        if default_layer_name in layer_names:
            return_value = default_layer_name
        else:
            return_value = _rs.AddLayer(default_layer_name)
        return return_value

    @classmethod                                ##  done 08-06
    def get_layer_name_from_user(cls):
        """Gets a valid name from the user. Returns:
            name            str. A unique and well-formed layer name
        """
        message_1 = "%s %s" % (
            "Enter the layer name.", 
            "It must be unique and contain no spaces or '#' characters")
        message_2 = "%s %s %s" % (
            "That name is already in use",
            "or it contains spaces or '#' characters.",
            "Try again")
        name = _rs.GetString(message_1)
        while not (
            cls._is_well_formed(name) and 
            cls._is_available(name)
        ):
            name = _rs.GetString(message_2)
        return name


    @classmethod                                ##  called
    def _is_well_formed(cls, name):
        """Receives:
            name            str. The name of a layer
        Returns:
            boolean         True, if the name is well-formed. False, 
                            otherwise
        """
        return_value = True
        for character in cls.prohibited_characters:
            if character in name:
                return_value = False
                break
        return return_value

    @classmethod                                ##  called
    def _is_available(cls, name):
        """Receives:
            name            str. The name of a layer
        Returns:
            boolean         True, if the name is available. False, otherwise
        """
        return_value = not _rs.IsLayer(name)
        return return_value

    @classmethod                                ##  called
    def get_frame_instance(cls, initial_shape):
        """Receives:
            initial_shape   str. The name of a layer containing one frame 
                            instance (i.e., an initial shape layer). The value 
                            is guaranteed
        Returns:
            frame_instance  guid. The guid of the frame instance on the layer 
                            Isn't this redundant?
        """
        if not _rs.IsLayer(initial_shape):
            message = "There is no layer named '%s'" % initial_shape
        all_frame_instances = _rs.BlockInstances(_s.Settings.frame_name)
        frame_instances_on_layer = []
        for frame_instance in all_frame_instances:
            if _rs.ObjectLayer(frame_instance) == initial_shape:
                frame_instances_on_layer.append(frame_instance)
        n_instances = len(frame_instances_on_layer)
        if n_instances == 0:
            message = "%s %s" % (
                "There is no frame instance", 
                "on the layer '%s'" % initial_shape)
            return_value = None
        elif n_instances == 1:
            message = None
            return_value = frame_instances_on_layer.pop()
        else:
            message = "%s %s" % (
                "There is more than 1 frame instance", 
                "on the layer '%s'" % initial_shape)
            return_value = None
        if message:
            print(message)
        return return_value

    @classmethod                                ##  called
    def get_frame_instance_pair(cls, rule):
        """Receives:
            rule            str. The name of a layer containing two frame 
                            instances (i.e., a rule layer). The value is 
                            guaranteed
        Returns:
            ordered_frame_instance_pair
                            (guid, guid). A pair of the guids of the two frame 
                            instances on the layer, ordered from left to right 
        """
        all_frame_instances = _rs.BlockInstances(_s.Settings.frame_name)
        frame_instances = []
        for instance_i in all_frame_instances:
            if _rs.ObjectLayer(instance_i) == rule:
                frame_instances.append(instance_i)
        p0 = _f.Frame.find_instance_position(frame_instances[0])
        p1 = _f.Frame.find_instance_position(frame_instances[1])
        if p0 < p1:
            ordered_frame_instance_pair = (
                frame_instances[0], frame_instances[1])
        elif p0 > p1:
            ordered_frame_instance_pair = (
                frame_instances[1], frame_instances[0])
        else:
            pass
        return ordered_frame_instance_pair

    @classmethod                                ##  called
    def contains_initial_shape(cls, name):
        """Receives:
            name            str. The name of the layer
        Returns:
            boolean         True, if the layer contains an initial shape. 
                            False, otherwise
        """
        value = cls._get_number_of_frames(name) == 1
        return value

    @classmethod                                ##  called
    def contains_rule(cls, name):
        """Receives:
            name            str. The name of the layer_name
        Returns:
            boolean         True, if the layer contains a rule. False, 
                            otherwise
        """
        value = cls._get_number_of_frames(name) == 2
        return value

    @classmethod                                ##  called
    def _get_number_of_frames(cls, layer_name):
        """Receives:
            layer_name      str. The name of the layer
        Returns:
            n               int. The number of frame instances on the layer
        """
        frame_name = _s.Settings.frame_name
        if not _rs.IsBlock(frame_name):
            n = 0
        else:
            frame_instance_guids = _rs.BlockInstances(frame_name)
            n = 0
            for guid in frame_instance_guids:
                if cls._contains_guid(guid, layer_name):
                    n = n + 1
        return n

    @classmethod                                ##  called
    def _contains_guid(cls, frame_guid, layer_name):
        """Receives:
            frame_guid      The guid of a frame instance
        Returns:
            value           boolean. True, if the layer contains the frame 
                            instance. False, otherwise
        """
        frame_layer_name = _rs.ObjectLayer(frame_guid)
        value = frame_layer_name == layer_name
        return value

    @classmethod                                ##  called
    def get_objects_on_layer(cls, frame_instance):
        """Receives:
            frame_instance  The guid of a frame instance
        Returns:
            objects_on_layer
                            [guid, ...]. A list of the guids of the objects 
                            on the layer containing the frame instance, if 
                            successful
            None            otherwise
        """
        layer_name = _rs.ObjectLayer(frame_instance)
        objects_on_layer = _rs.ObjectsByLayer(layer_name)
        return objects_on_layer

    @classmethod
    def find_next_layer_name(cls):              ##  make_shape_name
        """Finds the first available layer name of the appropriate form. 
        Starts from 'Shape_1'. Returns:
            layer_name_i    str. Has the form 'Shape_%i' % i, where i is an 
                            integer. False if unsuccessful
        """
        layer_names = _rs.LayerNames()
        i = 1
        while ('Shape_%i' % i) in layer_names:
            i += 1
        layer_name_i = 'Shape_%i' % i
        return layer_name_i

    @classmethod
    def new_shape_layer(cls):
        stem = 'Shape'
        name_in = Layer.make_layer_name(stem)
        name_out = Layer.new(name_in)
        return name_out

    @classmethod
    def new_rule_layer(cls, stem):
        """We expect the hash
        """
        # stem = 'Rule'
        # name_in = Layer.make_layer_name(stem)
        # name_out = Layer.new(name_in)
        name_out = Layer.new(stem)
        return name_out

    @classmethod
    def is_a_shape_layer(cls, layer_name):
        value = False
        if (type(layer_name) == str and
            layer_name[:5] == 'Shape'
        ):
            value = True
        return value

    @classmethod
    def make_layer_name(cls, stem_in):
        """Receives:
            stem_in         str
        Makes an available layer name by adding an index to the stem, 
        starting from 1. Returns:
            name_out        str. Name of the form 'stem i', i = 1 | 2 | ...
        """
        layer_names = _rs.LayerNames()
        i = 1
        name_out = '%s %i' % (stem_in, i)
        while name_out in layer_names:
            i += 1
            name_out = '%s %i' % (stem_in, i)
        return name_out




