from view import element as             _e
from view import layer as               _l
from view import py_line as             _pl
from view import py_point as            _pp
from view import rh_form as             _rf
import rhinoscriptsyntax as             _rs
from view import settings as            _s
from view import view_main as           _vm
from view import view_utilities as      _vu

color_c = _s.Settings.color_current_shape
color_d = _s.Settings.color_current_subshape
origin = _s.Settings.origin

class PyForm(object):
    """A PyForm is an object for translating between a Rhino form and an 
    agnostic spec. When created from a Rhino form (as a rule shape), it 
    incorporates a user-specified reference point. When created from an 
    agnostic spec, its 'reference point' is immaterial, as it is superceded 
    by the user-specified draw point. It contains only other Python objects - 
    PyPoints and PyLines. Possibly non-maximal. Immutable. 
    """
    class_name = 'PyForm'

    def __init__(self, py_points=[], py_lines=[]):
        """Receives:
            py_points       [PyPoint, ...]
            py_lines        [PyLine, ...]
        """
        method_name = '__init__'
        try:
            if not (
                _pp.PyPoint.is_a_list_of_py_points(py_points) and
                _pl.PyLine.is_a_list_of_py_lines(py_lines)
            ):
                raise TypeError
        except TypeError:
            message = '%s %s' % (
                'The arguments must be', 
                'a list of PyPoints and a list of PyLines')
            _vu._print_error_message(self.class_name, method_name, message)
            self.is_well_formed = False
        else:
            self.py_points = sorted(py_points)
            self.py_lines = sorted(py_lines)
            self.ag = self.agnostic = self._get_agnostic(
                self.py_points, self.py_lines)
            self.is_empty = self._is_empty()
            self.is_well_formed = True

    @classmethod                                ##  2017-10-28 20:11
    def new_from_objects_and_ref_point(cls, obs=[], ref_point=origin):
        """Receives:
            obs             [guid, ...]. A list of object guids
            ref_point       point3d. The reference point
        Returns:
            py_form         PyForm
        """
        text_dots, rh_lines = _e.Element.extract_text_dots_and_lines(obs)
        py_points = _pp.PyPoint.some_new_from_text_dots_and_ref_point(
            text_dots, ref_point)
        py_lines = _pl.PyLine.some_new_from_rh_lines_and_ref_point(
            rh_lines, ref_point)
        py_form = PyForm(py_points, py_lines)
        return py_form

    @classmethod
    def new_empty(cls):
        empty_py_form = PyForm([], [])
        return empty_py_form

    @classmethod
    def _get_agnostic(cls, py_points, py_lines):
        """Receives:
            py_points       [PyPoint, ...]
            py_lines        [PyLine, ...]
        Returns:
            agnostic        {'points': point_ags, 'line segments': line_ags}
                point_ags   [point_ag, ...]. Ordered
                line_ags    [line_ag, ...]. Ordered
                point_ag    (xyz, attrib_dict)
                line_ag     ((xyz1, xyz2), attrib_dict)
        """
        point_ags = [pp.ag for pp in py_points]
        line_ags = [pl.ag for pl in py_lines]
        agnostic = {
            'points': sorted(point_ags), 
            'line segments': sorted(line_ags), 
        }
        return agnostic

    def _is_empty(self):
        """Returns:
            value           boolean
        """
        value = False
        if (self.py_points == [] and
            self.py_lines == []
        ):
            value = True
        return value

    @classmethod
    def some_new_from_agnostics(cls, ag_shapes):
        """Receives:
            ag_shapes       [ag_shape, ...]
        Returns:
            py_forms        [PyForm, ...]. A list of corresponding PyForms
        """
        py_forms = [cls.new_from_agnostic(ag_shape) for ag_shape in ag_shapes]
        return py_forms

    @classmethod
    def new_from_agnostic(cls, ag_shape):
        """Receives:
            ag_shape        {'points': ag_points, 'line segments': ag_lines}
        Returns:
            py_form         PyForm. False if unsuccessful
        """
        ag_points = []
        ag_lines = []
        if 'points' in ag_shape:
            ag_points = ag_shape['points']
        if 'line segments' in ag_shape:
            ag_lines = ag_shape['line segments']
        py_points = _pp.PyPoint.some_new_from_agnostics(ag_points)
        py_lines = _pl.PyLine.some_new_from_agnostics(ag_lines)
        py_form = PyForm(py_points, py_lines)
        return py_form

    @classmethod
    def draw_many_nicely(cls, py_forms):
        """Receives:
            py_forms        [PyForm, ...]. Must not be empty. Guaranteed by 
                            caller
        Draws the PyForms in a row below and to the right of the origin, each 
        on a new, uniquely named layer and spaced according to the least 
        bounding box. Then moves existing shapes up. Reinstates the current 
        layer. Returns:
            forms_shifted   [guid_list, ...]. The forms arrayed, including 
                            empty forms. None if unsuccessful
        """
        method_name = 'draw_many_nicely'
        current_layer = _rs.CurrentLayer()
        forms_drawn = cls.draw_many(py_forms)
        if _rf.RhForm.are_empty(forms_drawn):
            message = "The next shapes are empty"
            print(message)
            forms_shifted = []
        else:
            draw_point_00, b_pts = cls._find_array_draw_point_and_bounding_box(
                forms_drawn)
            forms_arrayed = _rf.RhForm.array(forms_drawn, draw_point_00, b_pts)
            all_forms = _rf.RhForm.get_all()
            translation = cls._find_translation(b_pts)
            forms_shifted = cls._shift_forms(all_forms, translation)
            _rs.CurrentLayer(current_layer)
        return forms_shifted

    @classmethod
    def _find_array_draw_point_and_bounding_box(cls, forms_drawn):
        """Receives:
            forms_drawn     [PyForm, ...]. Must not be empty. Must not be 
                            [<empty form>]. Guaranteed by caller
        Returns:
            draw_point_00   Point3d. The draw point of the left most arrayed 
                            form
            b_pts           [Point3d, ...]. The least bounding box
        """
        b_pts = bounding_point_objects = _rf.RhForm.find_least_bounding_box(
            forms_drawn)
        p = upper_left_point = b_pts[3]
        bounding_box_side_y = _rs.VectorSubtract(b_pts[3], b_pts[0])
        gap_factor = 2
        draw_point_00 = _rs.VectorSubtract(
            origin, bounding_box_side_y * gap_factor)
        return (draw_point_00, b_pts)

    @classmethod
    def _find_translation(cls, b_pts):
        """Receives:
            b_pts           [Point3d, ...]. The points of the least bounding 
                            box
        Finds the translation for all forms. Returns:
            translation     Vector3d
        """
        bounding_box_side_y = _rs.PointSubtract(b_pts[3], b_pts[0])
        gap_factor = 2
        translation = _rs.VectorScale(bounding_box_side_y, gap_factor)
        return translation

    @classmethod
    def _shift_forms(cls, forms_in, translation):
        """Receives:
            forms_in        [RhForm, ...]. A list of forms before being shifted
            translation     Point3d
        Returns:
            forms_shifted   [RhForm, ...]. A list of forms after being shifted
        """
        forms_shifted = [
            _rs.MoveObjects(form, translation) for form in forms_in]
        return forms_shifted

    @classmethod
    def draw_many(cls, py_forms):
        """Receives:
            py_forms        [PyForm, ...]
        Draws the PyForms at their native locations (i.e., untranslated), on 
        new individual shape layers. Returns:
            forms           [guid_list, ...]
                guid_list   [guid, ...]
        """
        method_name = 'draw_many'
        forms = []
        for py_form in py_forms:
            layer_name = _l.Layer.new_shape_layer()
            result = py_form.draw(layer_name)
            if type(result) == None:
                message = "%s.%s. Is there a message from draw()?"
                _vu._print_error_message(cls.class_name, method_name, message)
            else:
                rh_form = result
                forms.append(rh_form.els)
        return forms

    def draw_as_group(self, layer_name, group_name):
        """Receives:
            layer_name      str. The name of an existing layer
            group_name      str. Name of an existing group
        Draws the form at its unmodified location on the layer, and as a 
        group. Returns:
            els             [guid, ...]. The guids of the elements drawn. 
                            None if unsuccessful
        """
        method_name = 'draw_as_group'
        if not type(group_name) == str:
            message = "The group name must be a string"
            _vu._print_error_message(self.class_name, method_name, message)
            return_value = None
        else:
            result = self.draw(layer_name)
            if result:
                els = result.els
                n_objects = _rs.AddObjectsToGroup(els, group_name)
                return_value = els
            else:
                return_value = None
        return return_value

    def draw(self, layer_name):
        """Receives:
            layer_name      str. The name of an existing layer
        Draws the form on the layer at its unmodified location, i.e., with 
        its origin at the display's origin. Returns:
            rh_form         RhForm. None if unsuccessful
        """
        method_name = 'draw'
        if not type(layer_name) == str:
            message = "The layer name must be a string"
            _vu._print_error_message(self.class_name, method_name, message)
            return_value = None
        elif not _rs.IsLayer(layer_name):
            message = "The layer name must exist"
            return_value = None
        else:
            elements = []
            for py_point in self.py_points:
                text_dot = py_point.draw(origin, layer_name)
                elements.append(text_dot)
            for py_line in self.py_lines:
                line = py_line.draw(origin, layer_name)
                elements.append(line)
            rh_form = _rf.RhForm(elements)
            return_value = rh_form
        return return_value

    def __hash__(self):
        py_point_hashes = [hash(pp) for pp in self.py_points]
        py_point_hash_tuple = tuple(py_point_hashes)
        py_points_hash = hash(py_point_hash_tuple)
        py_line_hashes = [hash(pl) for pl in self.py_lines]
        py_line_hash_tuple = tuple(py_line_hashes)
        py_lines_hash = hash(py_line_hash_tuple)
        value = hash((py_points_hash, py_lines_hash))
        return value

    def __eq__(self, other):
        value = False
        if (self.py_points == other.py_points and
            self.py_lines == other.py_lines
        ):
            value = True
        return value

    def __ne__(self, other):
        value = False
        if (self.py_points != other.py_points or
            self.py_lines != other.py_lines
        ):
            value = True
        return value

    def __str__(self):
        if self.py_points == []:
            pps_string = '[]'
        else:
            pp_strings = [str(pp) for pp in self.py_points]
            naked_pps_string = ', '.join(pp_strings)
            pps_string = '[%s]' % naked_pps_string
        if self.py_lines == []:
            pls_string = '[]'
        else:
            pl_strings = [str(pl) for pl in self.py_lines]
            naked_pls_string = ', '.join(pl_strings)
            pls_string = '[%s]' % naked_pls_string
        string = "(%s, %s)" % (pps_string, pls_string)
        return string

    def __repr__(self):
        pp_reprs = [repr(pp) for pp in self.py_points]
        pps_repr = ', '.join(pp_reprs)
        pl_reprs = [repr(pl) for pl in self.py_lines]
        pls_repr = ', '.join(pl_reprs)
        string = "_pf.PyForm([%s], [%s])" % (pps_repr, pls_repr)
        return string


