from view import agnostic as            _a
import                                  copy
from view import py_point as            _pp
import rhinoscriptsyntax as             _rs
from view import settings as            _s
from view import view_utilities as      _vu

eeq_xyz = _vu.eeq_xyz
ege_xyz = _vu.ege_xyz
egt_xyz = _vu.egt_xyz
ele_xyz = _vu.ele_xyz
elt_xyz = _vu.elt_xyz
ene_xyz = _vu.ene_xyz
origin = _s.Settings.origin

class PyLine(object):
    """Contains the coordinates and attributes of a line. See PyForm
    """

    def __init__(
        self, xyz1, xyz2, labels_in=[], weight=_s.Settings.default_line_weight
    ):
        """Receives:
            xyz1            (num, num, num). The coordinates of point 1
            xyz2            (num, num, num). The coordinates of point 2. Must 
                            be different from xyz1
            labels_in       str or [str, ...]
            weight          num. Greater than 0
        It is absolute; i.e., it already takes into account the line's 
        reference point. The attributes are implemented on Rudi's side; only 
        labels are implemented in Rhino. Immutable. 
        """
        class_name = 'PyLine'
        method_name = '__init__'
        # if not labels_in:
        #     labels_in = []
        try:
            if not(
                type(xyz1) == tuple and
                type(xyz2) == tuple and
                len(xyz1) == 3 and
                len(xyz2) == 3 and
                _vu.is_a_number(xyz1[0]) and
                _vu.is_a_number(xyz1[1]) and
                _vu.is_a_number(xyz1[2]) and
                _vu.is_a_number(xyz2[0]) and
                _vu.is_a_number(xyz2[1]) and
                _vu.is_a_number(xyz2[2]) and
                # (   type(labels_in) == str or
                #     _vu.is_a_list_of_strings(labels_in)) and
                _vu.is_a_number(weight)
            ):
                raise TypeError
            if (eeq_xyz(xyz1, xyz2) or
                weight <= 0
            ):
                raise ValueError
        except TypeError:
            message = "%s %s %s %s" % (
                "Bad type.", 
                "The points must be triples of numbers.", 
                "The label(s) must be string(s).", 
                "The weight must be a number.", 
            )
            _vu._print_error_message(class_name, method_name, message)
            self.is_well_formed = False
        except ValueError:
            message = '%s %s' % (
                "The points must be distinct.", 
                "The weight must be greater than 0.", 
            )
            _vu._print_error_message(class_name, method_name, message)
            self.is_well_formed = False
        else:
            pp1 = _pp.PyPoint.new_from_xyz_and_labels(xyz1)
            pp2 = _pp.PyPoint.new_from_xyz_and_labels(xyz2)
            if pp1 < pp2:
                self.tail = self.xyz1 = xyz1
                self.head = self.xyz2 = xyz2
            else:
                self.tail = self.xyz1 = xyz2
                self.head = self.xyz2 = xyz1
            xyz_pair = (self.tail, self.head)
            if labels_in == []:
                self.labels = []
            elif labels_in == '':
                self.labels = []
            elif type(labels_in) == str:
                label = labels_in
                self.labels = [label]
            else:
                unique_label_set = set(labels_in)
                unsorted_unique_labels = list(unique_label_set)
                self.labels = sorted(unsorted_unique_labels)
            self.labels_str = '#'.join(self.labels)
            self.weight = weight                ##  set print weight
            self.ag = self.agnostic = _a.Agnostic.new_ag_line(
                self.tail, self.head, self.labels, self.weight)
            self.is_well_formed = True

    def _contains_duplicate_elements(self, elements):
        """Receives:
            elements        [element, ...]. Possibly empty
        Returns:
            value           boolean. True if one or more elements is 
                            duplicated. False otherwise
        """
        value = False
        elements_copy = copy.copy(elements)
        element_set = set(elements_copy)
        if len(elements) != len(element_set):
            value = True
        return value

    def _is_a_list_of_strings(self, item):
        """Receives:
            item
        Returns:
            value           boolean. True if the item is a list (possibly 
                            empty) of strings. False otherwise
        """
        value = False
        if type(item) == list:
            elements = item
            for el in elements:
                if not type(el) == str:
                    break
            value = True
        return value

    @classmethod
    def some_new_from_rh_lines_and_ref_point(cls, lines, ref_point=origin):
        """Receives:
            lines           [guid, ...]. A list of line guids
            ref_point       point3d
        Returns:
            py_lines        [PyLine, ...]. A list of PyLine objects relative 
                            to the reference point
        """
        py_lines = []
        for line in lines:
            py_line = PyLine.new_from_rh_line_and_ref_point(line, ref_point)
            py_lines.append(py_line)
        return py_lines

    @classmethod
    def new_from_rh_line_and_ref_point(cls, line, ref_point=origin):
        """Receives:
            line            guid. A line guid   ##  polyline?
            ref_point       point3d
        Returns:
            py_line         PyLine. A PyLine object relative to the reference 
                            point
        """
        p1 = _rs.CurveStartPoint(line)
        p2 = _rs.CurveEndPoint(line)
        p10 = _rs.PointSubtract(p1, ref_point)
        p20 = _rs.PointSubtract(p2, ref_point)
        xyz1, xyz2 = tuple(p10), tuple(p20)
        label = _rs.ObjectName(line)
        if label == '':
            labels = []
        else:
            labels = [label]
        py_line = PyLine(xyz1, xyz2, labels)
        return py_line

    @classmethod
    def is_a_list_of_py_lines(cls, item):
        """Receives:
            item            any type
        Returns:
            value           boolean. True if item is a list (possibly empty) 
                            of PyLine objects. False otherwise
        """
        value = True
        if not type(item) == list:
            value = False
        else:
            from view import py_line
            elements = item
            for element in elements:
                if not type(element) == PyLine:
                    value = False
                    break
        return value

    @classmethod
    def some_new_from_agnostics(cls, ag_lines):
        """Receives:
            ag_lines        [ag_line, ...]
                ag_line     ((xyz, xyz), attribute_dict)
        Returns:
            py_lines        [PyLine, ...]
        """
        py_lines = []
        for ag_line in ag_lines:
            py_line = PyLine.new_from_agnostic(ag_line)
            py_lines.append(py_line)
        return py_lines

    @classmethod
    def new_from_agnostic(cls, ag_line):
        """Receives:
            ag_line         (xyz_pair, attribute_dict)
                xyz         (num, num, num)
                attribute_dict
                            {'labels': [label, ...], 'weight': num, ...}
        Returns:
            py_line         PyLine
        """
        xyz_pair, attrib_dict = ag_line
        xyz1, xyz2 = xyz_pair
        if 'labels' in attrib_dict:
            labels = attrib_dict['labels']
        else:
            labels = []
        if 'weight' in attrib_dict:
            weight = attrib_dict['weight']
        else:
            weight = _s.Settings.default_line_weight
        py_line = PyLine(xyz1, xyz2, labels, weight)
        return py_line

    def draw(self, ref_point, layer_name):
        """Receives:
            ref_point       point3d
            layer_name      str. The name of an existing layer
        Draws the line relative to the reference point on the layer. Returns:
            line            guid. False if unsuccessful
        """
        if not _rs.CurrentLayer() == layer_name:
            _rs.CurrentLayer(layer_name)
        offset_tail = _rs.PointAdd(self.tail, ref_point)
        offset_head = _rs.PointAdd(self.head, ref_point)
        line = _rs.AddLine(offset_tail, offset_head)
        if line:
            _rs.ObjectName(line, self.labels_str)
        return line

    def __hash__(self):
        """Returns:
            value           int
        """
        n = _s.Settings.n_round_digits
        xyz1_list_rounded = [round(coord, n) for coord in self.xyz1]
        xyz2_list_rounded = [round(coord, n) for coord in self.xyz2]
        xyz1_tuple_rounded = tuple(xyz1_list_rounded)
        xyz2_tuple_rounded = tuple(xyz2_list_rounded)
        labels_tuple = tuple(self.labels)
        value = hash(
            (   xyz1_tuple_rounded, 
                xyz2_tuple_rounded, 
                labels_tuple, 
                self.weight))
        return value

    def __eq__(self, other):
        if not type(other) == PyLine:
            value = None
        elif (
            eeq_xyz(self.tail, other.tail) and
            eeq_xyz(self.head, other.head) and
            self.labels == other.labels and
            self.weight == other.weight
        ):
            value = True
        else:
            value = False
        return value

    def __ge__(self, other):
        value = False
        if not type(other) == PyLine:
            value = False
        elif egt_xyz(self.tail, other.tail):
            value = True
        elif eeq_xyz(self.tail, other.tail):
            if egt_xyz(self.head, other.head):
                value = True
            elif eeq_xyz(self.head, other.head):
                if self.labels > other.labels:
                    value = True
                elif self.labels == other.labels:
                    if self.weight >= other.weight:
                        value = True
        return value

    def __gt__(self, other):
        value = False
        if not type(other) == PyLine:
            value = None
        elif egt_xyz(self.tail, other.tail):
            value = True
        elif eeq_xyz(self.tail, other.tail):
            if egt_xyz(self.head, other.head):
                value = True
            elif eeq_xyz(self.head, other.head):
                if self.labels > other.labels:
                    value = True
                elif self.labels == other.labels:
                    if self.weight > other.weight:
                        value = True
        return value

    def __le__(self, other):
        value = False
        if not type(other) == PyLine:
            value = None
        elif elt_xyz(self.tail, other.tail):
            value = True
        elif eeq_xyz(self.tail, other.tail):
            if elt_xyz(self.head, other.head):
                value = True
            elif eeq_xyz(self.head, other.head):
                if self.labels < other.labels:
                    value = True
                elif self.labels == other.labels:
                    if self.weight <= other.weight:
                        value = True
        return value

    def __lt__(self, other):
        value = False
        if not type(other) == PyLine:
            value = None
        elif elt_xyz(self.tail, other.tail):
            value = True
        elif eeq_xyz(self.tail, other.tail):
            if elt_xyz(self.head, other.head):
                value = True
            elif eeq_xyz(self.head, other.head):
                if self.labels < other.labels:
                    value = True
                elif self.labels == other.labels:
                    if self.weight < other.weight:
                        value = True
        return value

    def __ne__(self, other):
        value = False
        if not type(other) == PyLine:
            value = None
        elif (
            ene_xyz(self.tail, other.tail) or
            ene_xyz(self.head, other.head) or
            self.labels != other.labels or
            self.weight != other.weight
        ):
            value = True
        return value

    def __str__(self):
        string = '(%s, %s, %s, %s)' % (
            str(self.tail), 
            str(self.head), 
            self.labels, 
            self.weight, 
        )
        return string

    def __repr__(self):
        string = '_pl.PyLine(%s, %s, %s, %s)' % (
            str(self.xyz1), 
            str(self.xyz2), 
            self.labels, 
            self.weight, 
        )
        return string


