from view import agnostic as            _a
import rhinoscriptsyntax as             _rs
from view import settings as            _s
from view import text_dot as            _td
from view import view_utilities as      _vu

eeq = _vu.eeq
eeq_xyz = _vu.eeq_xyz
egt_xyz = _vu.egt_xyz
elt_xyz = _vu.elt_xyz
origin = _s.Settings.origin

class PyPoint(object):
    """Contains the coordinates and attributes of a point. See PyForm. 
    Immutable
    """

    def __init__(self, x, y, z, labels_in=[]):
        """Receives:
            x               num
            y               num
            z               num
            labels_in       str or [str, ...]
        Immutable
        """
        class_name = 'PyPoint'
        method_name = '__init__'
        try:
            if not (
                _vu.is_a_number(x) and
                _vu.is_a_number(y) and
                _vu.is_a_number(z) and
                (   type(labels_in) == str or
                    _vu.is_a_list_of_strings(labels_in))
            ):
                raise TypeError
        except TypeError:
            message = '%s %s' % (
                "The arguments must be 3 numbers and optionally", 
                "either a string or a list of strings")
            _vu._print_error_message(class_name, method_name, message)
            self.is_well_formed = False
        else:
            self.x = x
            self.y = y
            self.z = z
            self.xyz = (x, y, z)
            if type(labels_in) == str:
                if labels_in == '':
                    self.labels = []
                else:
                    self.labels = [labels_in]
            else:
                unique_label_set = set(labels_in)
                unsorted_unique_labels = list(unique_label_set)
                self.labels = sorted(unsorted_unique_labels)
            self.labels_str = '#'.join(self.labels)
            self.ag = self.agnostic = _a.Agnostic.new_ag_point(
                self.x, self.y, self.z, self.labels)
            self.is_well_formed = True

    @classmethod
    def some_new_from_xyz_label_pairs(cls, xyz_label_pairs):
        """Receives:
            xyz_label_pairs (xyz, labels)
                xyz         (num, num, num)
                labels      str or [str, ...]
        """
        py_points = []
        for pair in xyz_label_pairs:
            xyz, labels = pair
            x, y, z = xyz
            pp = PyPoint(x, y, z, labels)
            py_points.append(pp)
        return py_points

    @classmethod
    def new_from_xyz_and_labels(cls, xyz, labels=[]):
        """Receives:
            xyz             (num, num, num)
            labels          str or [str, ...]
        Returns:
            pp              PyPoint
        """
        x, y, z = xyz
        pp = PyPoint(x, y, z, labels)
        return pp

    @classmethod
    def some_new_from_text_dots_and_ref_point(cls, text_dots, ref_point):
        """Receives:
            text_dots       [guid, ...]. A list of text dot guids
            ref_point       point3d. The reference point
        Ignores guids of non-text dots. Returns:
            py_points       [PyPoint, ...]. A list of PyPoint objects
                            relative to the reference point
        """
        py_points = []
        for td in text_dots:
            pp = cls.new_from_text_dot_and_ref_point(td, ref_point)
            if pp:
                py_points.append(pp)
        return py_points

    @classmethod
    def new_from_text_dot_and_ref_point(cls, text_dot, ref_point):
        """Receives:
            text_dot        guid
            ref_point       point3d
        Returns:
            py_point        PyPoint. Relative to the reference point. If 
                            successful. None otherwise
        """
        method_name = 'new_from_text_dot_and_ref_point'
        try:
            if not (
                _td.TextDot.is_text_dot(text_dot) and
                len(ref_point) == 3
            ):
                raise TypeError
        except TypeError:
            message = '%s: %s' % (
                method_name, 
                "The arguments must be a text dot and a point3d")
            print(message)
            return_value = None
        else:
            td = text_dot                           ##  guid
            p_rel = _rs.TextDotPoint(td)            ##  p3d
            p_abs = _rs.PointSubtract(p_rel, ref_point) ##  p3d
            pp_x, pp_y, pp_z = tuple(p_abs)
            label = _rs.TextDotText(td)
            pp = py_point = PyPoint(pp_x, pp_y, pp_z, label)
            return_value = py_point
        return return_value

    @classmethod
    def some_new_from_agnostics(cls, ag_points):
        """Receives:
            ag_points       [ag_point, ...]
                ag_point    (xyz, attribute_dict)
        Returns:
            py_points       [PyPoint, ...]. Relative to the reference point
        """
        py_points = [PyPoint.new_from_agnostic(ag) for ag in ag_points]
        return py_points

    @classmethod
    def new_from_agnostic(cls, ag_point):
        """Receives:
            ag_point        (xyz, attribute_dict)
                xyz         (num, num, num)
                attribute_dict
                            {'labels': [label, ...], ...}
        Returns:
            py_point        PyPoint. The corresponding PyPoint
        """
        xyz, attrib_dict = ag_point
        if 'labels' in attrib_dict:
            labels = attrib_dict['labels']
        else:
            labels = []
        py_point = PyPoint.new_from_xyz_and_labels(xyz, labels)
        return py_point

    @classmethod
    def is_a_list_of_py_points(cls, item):
        """Receives:
            item            any type
        Returns:
            value           boolean. True if item is a list (possibly empty) 
                            of PyPoint objects. False otherwise
        """
        value = True
        if not type(item) == list:
            value = False
        else:
            import py_point
            elements = item
            for element in elements:
                if not type(element) == PyPoint:
                    value = False
                    break
        return value

    def draw(self, ref_point, layer_name):
        """Receives:
            ref_point       point3d
            layer_name      str. The name of an existing layer
        Draws the point relative to the reference point on the layer. Returns: 
            text_dot        guid. False if unsuccessful
        """
        if not _rs.CurrentLayer() == layer_name:
            _rs.CurrentLayer(layer_name)
        offset_xyz = _rs.PointAdd(self.xyz, ref_point)
        text_dot = _rs.AddTextDot(self.labels_str, offset_xyz)
        return text_dot

    def __hash__(self):
        """Returns:
            value           int
        """
        n = _s.Settings.n_round_digits
        xyz_list_rounded = [round(coord, n) for coord in self.xyz]
        xyz_tuple_rounded = tuple(xyz_list_rounded)
        label_tuple = tuple(self.labels)
        labels_hash = hash(label_tuple)
        value = hash((xyz_tuple_rounded, labels_hash))
        return value

    def __eq__(self, other):
        if not type(other) == PyPoint:
            value = None
        else:
            value = (
                eeq_xyz(self.xyz, other.xyz) and
                self.labels == other.labels)
        return value

    def __ne__(self, other):
        value = False
        if not type(other) == PyPoint:
            value = None
        else:
            value = (
                not eeq_xyz(self.xyz, other.xyz) or
                self.labels != other.labels)
        return value

    def __ge__(self, other):
        value = False
        if not type(other) == PyPoint:
            value = None
        elif egt_xyz(self.xyz, other.xyz):
            value = True
        elif eeq_xyz(self.xyz, other.xyz):
            if self.labels >= other.labels:
                value = True
        return value

    def __gt__(self, other):
        value = False
        if not type(other) == PyPoint:
            value = None
        if egt_xyz(self.xyz, other.xyz):
            value = True
        elif (
            eeq_xyz(self.xyz, other.xyz) and
            self.labels > other.labels
        ):
            value = True
        return value

    def __le__(self, other):
        value = False
        if not type(other) == PyPoint:
            value = None
        elif elt_xyz(self.xyz, other.xyz):
            value = True
        elif eeq_xyz(self.xyz, other.xyz):
            if self.labels <= other.labels:
                value = True
        return value

    def __lt__(self, other):
        value = False
        if not type(other) == PyPoint:
            value = None
        elif elt_xyz(self.xyz, other.xyz):
            value = True
        elif (eeq_xyz(self.xyz, other.xyz) and
            self.labels < other.labels
        ):
            value = True
        return value

    def __str__(self):
        string = "((%s, %s, %s), %s)" % (self.x, self.y, self.z, self.labels)
        return string

    def __repr__(self):
        string = "_pp.PyPoint(%s, %s, %s, %s)" % (
            self.x, self.y, self.z, self.labels)
        return string

