import                                          copy
from view import agnostic as                    _ag
from view import arrow as                       _ar
# from controller import controller_main as       _cm
from view import element as                     _e
from view import frame as                       _fr
from view import grammar as                     _g
from view import layer as                       _l
from view import py_form as                     _pf
from view import rh_form as                     _rf
from view import rh_rule as                     _rr
import rhinoscriptsyntax as                     _rs
from view import settings as                    _s
from view import view_main as                   _vm
from view import view_utilities as              _vu

color_a = _s.Settings.color_left
color_b = _s.Settings.color_right

class PyRule(object):
    class_name = 'PyRule'

    def __init__(self, py_a, py_b, descr=None):
        """Receives:
            py_a            PyForm. The left py_form. Must not be empty
            py_b            PyForm. The right py_form. May be empty
            descr           str
        Created from Rhino forms. Not received from sortal. Calculates and 
        remembers its id (an int). (The name (a str) is derived from the id 
        when the rule is drawn and retained as the layer name.) Immutable
        """
        class_name = PyRule.class_name
        method_name = '__init__'
        try:
            if not (
                type(py_a) == _pf.PyForm and
                type(py_b) == _pf.PyForm
            ):
                raise TypeError
            if py_a.is_empty:
                raise ValueError
            if not (
                type(descr) == str or
                descr == None
            ):
                raise TypeError
        except TypeError:
            message = '%s %s' % (
                "The arguments must be PyForm objects", 
                "and an optional string")
            _vu._print_error_message(class_name, method_name, message)
            self.is_well_formed = False
        except ValueError:
            message = "The left form must not be empty"
            _vu._print_error_message(class_name, method_name, message)
            self.is_well_formed = False
        else:
            self.a = py_a
            self.b = py_b
            self.id_ = hash(self)
            if descr == None:
                self.descr = ''
            else:
                self.descr = descr
            self.ag = self.agnostic = _ag.Agnostic.new_ag_rule( # who needs this?
                self.a.ag, self.b.ag, self.id_, self.descr)
            self.is_well_formed = True

    @classmethod
    def many_new_from_rule_groups(cls, rule_layer_names):
        """Receives:
            rule_layer_names
                            [str, ...]. A list of names of layers containing 
                            rules. Guaranteed
        Returns:
            py_rules        [PyRule, ...]. A list of PyRules
        No test
        """
        py_rules = []
        for name in rule_layer_names:
            py_rule = PyRule.new_from_rule_group(name)
            py_rules.append(py_rule)
        return py_rules

    @classmethod
    def new_from_rule_group(cls, rule_layer_name):
        """Receives:
            rule_layer_name str. The name of a layer containing a rule. 
                            Guaranteed
        Returns:
            py_rule         PyRule
        """
        method_name = 'new_from_rule_group'
        return_value = None
        if not type(rule_layer_name) == str:
            error_message = "The layer name must be a string"
            print('%s. %s: %s' % (
                cls.class_name, 
                method_name, 
                error_message))
        else:
            obs = _rs.ObjectsByLayer(rule_layer_name)
            if (obs == None or
                obs == []
            ):
                error_message = "There is no layer with the name %s" % (
                    rule_layer_name)
                print('%s.%s: %s' % (
                    cls.class_name, 
                    method_name, 
                    error_message))
            else:
                form_a, form_b = _rf.RhForm.get_form_pair_from_rule_objects(obs)
                pf_a = _pf.PyForm.new_from_objects_and_ref_point(
                    form_a.els, form_a.ref_point)
                pf_b = _pf.PyForm.new_from_objects_and_ref_point(
                    form_b.els, form_b.ref_point)
                py_rule = PyRule(pf_a, pf_b)
                return_value = py_rule
        return return_value

    def _draw_grouped_forms(self, layer_name):
        """Receives:
            layer_name      str
        Draws and groups the left and right forms at their unmodified 
        locations. Returns:
            group_a_name    str. The name of the left form group
            group_b_name    str. The name of the right form group
        """
        group_a_name_in, group_b_name_in = (
            self._make_form_group_names(layer_name))
        group_a_name = _rs.AddGroup(group_a_name_in)
        group_b_name = _rs.AddGroup(group_b_name_in)
        objects_a = self.a.draw_as_group(layer_name, group_a_name)
        objects_b = self.b.draw_as_group(layer_name, group_b_name)
        return (group_a_name, group_b_name)

    def _make_form_group_names(self, layer_name):
        """Receives:
            layer_name      str
        Returns:
            group_a_name    str. The name of the left form group
            group_b_name    str. The name of the right form group
        """
        group_a_name = '%s%s%s' % (
            layer_name, 
            _s.Settings.group_name_separator, 
            _s.Settings.group_name_suffix_left)
        group_b_name = '%s%s%s' % (
            layer_name, 
            _s.Settings.group_name_separator, 
            _s.Settings.group_name_suffix_right)
        return (group_a_name, group_b_name)

    def _find_common_bounding_points(self, objects_a, objects_b):
        """Receives:
            objects_a       [guid, ...]. A list of the guids of the objects 
                            in the left group
            objects_b       [guid, ...]. A list of the guids of the objects 
                            in the right group
        Returns:
            common_bounding_points
                            [p3d, ...]. A list of 8 3d points defining the 
                            bounding box of the left and right object groups. 
                            An empty list if both object groups are empty. 
                            None if unsuccessful
        """
        objects_ab = copy.copy(objects_a)
        objects_ab.extend(objects_b)
        n_objects = _rs.SelectObjects(objects_ab)
        if n_objects == 0:
            common_bounding_points = []
        else:
            common_bounding_points = _rs.BoundingBox(objects_ab)
        return common_bounding_points

    def _find_translation_to_staging_area(self, insert_point, scale):
        """Receives:
            insert_point    point3d. The insertion point of the padded frame
            scale           (num, num, num). The scale of the padded frame. 
                            num >= 0
        Returns:
            t               vector3d. The translation to the staging area
        """
        origin_p3d = _s.Settings.origin_p3d
        v1 = start_to_origin = _rs.VectorCreate(origin_p3d, insert_point)
        rule_margin = _s.Settings.rule_margin
        finish_x = 0 + rule_margin[0] - scale[0]
        finish_y = 0 + rule_margin[1]
        finish_z = 0 + rule_margin[2]
        finish = _vu.get_point3d((finish_x, finish_y, finish_z))
        v2 = origin_to_finish = _rs.VectorCreate(finish, origin_p3d)
        t = _rs.VectorAdd(v1, v2)
        return t

    def _find_translation_of_left_shape(self, scale):
        """Receives:
            scale           (num, num, num). The scale of the padded frame
        Returns:
            t               vector3d. The translation from the right shape to 
                            the left shape
        """
        v = (scale[0], 0, 0)
        v_scale = _s.Settings.rule_left_shape_offset_factor
        t = _rs.VectorScale(v, v_scale)
        return t

    def _find_arrow_position(self, insert_point, scale):
        """Receives:
            insert_point    point3d. The insertion point of the right padded 
                            frame
            scale           (num, num, num). The scale of the padded frame
        Returns:
            p_arrow         point3d. The insertion point of the arrow
        """
        t = (
            scale[0] * _s.Settings.rule_arrow_offset_factor_x, 
            scale[1] * _s.Settings.rule_arrow_offset_factor_y, 
            0)
        p_arrow = _rs.PointAdd(insert_point, t)
        print('insert_point: %s' % str(insert_point))
        print('p_arrow: %s' % str(p_arrow))
        return p_arrow

    def _lay_out_rule(self):                    ##  2018-05-28 06:59
        """Receives:
            framed_group_a
            framed_group_b
        Lays out rule parts: left shape, arrow, right shape. 
        """
        pass

    @classmethod
    def exists(cls, py_rule):
        """Receives:
            py_rule         PyRule
        Returns:
            value           boolean. True if a rule layer with the same name 
                            stem (i.e., not the disambiguating extension) 
                            exists. False otherwise
        """
        value = False
        rule_name = str(py_rule.id_)
        layer_names = _vm.ViewMain.get_rule_layer_names()
        for layer_name in layer_names:
            layer_name_stem = cls._get_stem(layer_name)
            if rule_name == layer_name_stem:
                value = True
                break
        return value

    @classmethod
    def _get_stem(cls, layer_name):
        """Receives:
            layer_name      str. Type guaranteed
        Removes the disambiguating extension (i.e., separator plus integer) 
        from layer_name. Returns:
            layer_name_stem str
        """
        if not type(layer_name) == str:
            layer_name_stem = None
        else:
            separator = '_'
            if separator in layer_name:
                i = layer_name.index(separator)
                layer_name_stem = layer_name[:i]
            else:
                layer_name_stem = layer_name
        return layer_name_stem

    def draw(self):
        """Creates a new layer whose name is derived from the id and 
        disambiguated as necessary. Draws the forms and the arrow in the 
        display zone on the new layer. Groups the elements (i.e., no form 
        groups) and the arrow under the id_name. Uses RhForm and RhRule 
        objects. Returns:
            rh_rule         RhRule. The drawn version of the py_rule
        """
        if not self.is_well_formed:
            return_value = None
        else:
            layer_name = self._make_unique_name()
            _rs.AddLayer(layer_name)
            rh_form_a = self.a.draw(layer_name)
            rh_form_b = self.b.draw(layer_name)
            rh_rule = _rr.RhRule(rh_form_a, rh_form_b, layer_name)
            rh_form_b_frame_pos = rh_rule.arrange()
            return_value = rh_rule
        return return_value

    def _make_unique_name(self):
        """Creates a unique name based on the id. Returns:
            name            str
        """
        layer_names = _rs.LayerNames()
        id_str = str(self.id_)
        if id_str not in layer_names:
            name = id_str
        else:
            name = copy.copy(id_str)
            separator = '_'
            if separator in name:
                i = name.index(separator)
                stem = name[:i]
                suffix_in = name[(i + 1):]
            else:
                stem = name
                suffix_in = '0'
            suffix_in_int = int(suffix_in)
            suffix_out_int = suffix_in_int + 1
            name = '%s_%i' % (stem, suffix_out_int)
            while name in layer_names:
                suffix_out_int += 1
                name = '%s_%i' % (stem, suffix_out_int)
        return name

    def __hash__(self):
        hash_a = hash(self.a)
        hash_b = hash(self.b)
        value = hash((hash_a, hash_b))
        return value

    def __eq__(self, other):
        value = False
        if (self.a == other.a and
            self.b == other.b
        ):
            value = True
        return value

    def __ne__(self, other):
        value = False
        if (self.a != other.a or
            self.b != other.b
        ):
            value = True
        return value

    def __str__(self):
        string = '(%s, %s, %i, %s)' % (
            str(self.a), str(self.b), self.id_, self.descr)
        return string

    def __repr__(self):
        string = '_pr.PyRule(%s, %s, %s)' % (
            repr(self.a), 
            repr(self.b), 
            repr(self.descr))

        return string


