from view import element as             _e
from view import frame as               _fr
from view import layer as               _l
from view import py_point as            _pp
import rhinoscriptsyntax as             _rs
from view import settings as            _s
from view import text_dot as            _td

color_a = _s.Settings.color_left
color_b = _s.Settings.color_right
color_c = _s.Settings.color_current_shape
color_d = _s.Settings.color_current_subshape
origin_p3d = _s.Settings.origin_p3d

class RhForm(object):
    """A container for located Rhino objects, i.e., objects plus a reference 
    point. Has a set. 
    """
    def __init__(self, obs=[], ref_point=origin_p3d):
        """Receives:
            obs                 [guid, ...]. A list of object guids
            ref_point           point3D or (num, num, num)
        Filters out non-elements; i.e., retains only lines and text dots
        """
        method_name = '__init__'
        try:
            if not (
                type(obs) == list and
                _td.TextDot.is_a_num_triple(ref_point)
            ):
                raise TypeError
        except TypeError:
            message = "%s %s" % (
                "The arguments must be: 1) a list of Rhino objects", 
                "and 2) a point3d"
            )
            self._print_error_message(method_name, message)
            self.is_well_formed = False
        else:
            els = _e.Element.extract_from_objects(obs)
            self.els = sorted(els)
            if els == []:
                self.ref_point = origin_p3d
            else:
                self.ref_point = ref_point
            self.frame = None
            self.is_well_formed = True

    @classmethod
    def new_empty(cls):
        empty_form = RhForm()
        return empty_form

    @classmethod
    def get_form_pair_from_rule_objects(cls, obs):
        """Receives:
            obs             [guid, ...]. A list of object guids. Must include 
                            exactly 2 non-overlapping (in the x- and xy-
                            dimensions) frame instances. Must include at least 
                            1 object. May include non-elements
        Returns:
            form_a          RhForm. The left form. Must not be empty. 
            form_b          RhForm. The right form
            None            if unsuccessful
        """
        return_value = None
        if (obs == None or
            len(obs) == 0
        ):
            error_message = "There must be points or lines"
            print(error_message)
        else:
            els, frames = _e.Element.extract_with_frames_from_objects(obs)
            if len(els) == 0:
                error_message = "There are no elements"
                print(error_message)
            elif (len(frames) != 2 or
                _fr.Frame.overlaps_object(frames[0], frames[1])
            ):
                error_message = "There should be 2 non-overlapping frames"
                print(error_message)
            else:
                ff = ordered_frames = _fr.Frame.order_pair(frames)
                els_a, els_b = el_list_pair = (
                    _e.Element.extract_by_frame(els, ff))
                ref_pt_a, ref_pt_b = ref_points = [
                    _fr.Frame.find_instance_position(f) for f in ff]
                form_a = RhForm(els_a, ref_pt_a)
                form_b = RhForm(els_b, ref_pt_b)
                return_value = (form_a, form_b)
        return return_value

    @classmethod
    def is_a_form(self, obs):
        """Receives:
            obs             [guid, ...]
        Identifies a list of objects (not a RhForm object) as distinct from a 
        rule. Returns:
            value           boolean. True if there is at least one element in 
                            obs. False otherwise
        """
        value = False
        for ob in obs:
            if _e.Element.is_element(ob):
                value = True
                break
        return value

    @classmethod
    def are_empty(cls, forms):
        """Receives:
            forms           [guid_list, ...]
        Returns:
            value           boolean. True if all forms are empty. False 
                            otherwise
        """
        value = True
        for f in forms:
            if not f == []:
                value = False
                break
        return value

    @classmethod
    def find_least_bounding_box(cls, forms):
        """Receives:
            forms           [ob_list, ...]. A list of guid lists, not of 
                            RhForm objects
                ob_list     [guid, ...]
        Finds the least common bounding box for the forms. Returns:
            point_obs       [point3d, ...]. The 8 vertices of the least 
                            bounding box. None if unsuccessful
        """
        objects = []
        for form in forms:
            objects.extend(form)
        if objects == []:
            return_value = None
        else:
            return_value = point_obs = _rs.BoundingBox(objects)
        return return_value

    @classmethod
    def array(cls, forms, draw_point, bounding_points):
        """Receives:
            forms           [ob_list, ...]
                ob_list     [guid, ...]
            draw_point      xyz or point3d. The lower left corner of the array
            bounding_points [point3d, ...]. A list of the 8 vertices of the 
                            least bounding box
        Moves the forms into a 1-dimensional array, at a spacing determined 
        by the least bounding box. Returns:
            forms_arrayed   [ob_list, ...]. The guid lists after being 
                            arrayed. None if unsuccessful
        """
        b_pts = bounding_points
        if (b_pts == None or
            b_pts == []
        ):
            return_value = None
        else:
            forms_arrayed = []
            start_point_to_origin = _rs.PointSubtract(origin_p3d, b_pts[0])
            # print('start_point: %s' % str(b_pts[0]))
            origin_to_array_00 = draw_point
            # print('draw_point: %s' % str(draw_point))
            start_point_to_array_00 = _rs.VectorAdd(
                start_point_to_origin, 
                origin_to_array_00)
            # print('start_point_to_array_00: %s' % str(start_point_to_array_00))
            bounding_box_side_x = _rs.PointSubtract(b_pts[1], b_pts[0])
            gap_factor = 2
            array_00_to_array_10 = bounding_box_side_x * gap_factor
            for i in range(len(forms)):
                array_00_to_array_i0 = array_00_to_array_10 * i
                translation = _rs.VectorAdd(
                    start_point_to_array_00, 
                    array_00_to_array_i0)
                form = forms[i]
                form_arrayed = _rs.MoveObjects(form, translation)
                forms_arrayed.append(form_arrayed)
            return_value = forms_arrayed
        return return_value

    @classmethod
    def get_all(cls):
        """Returns:
            forms           [guid_list, ...]. A list of guid lists, one per 
                            shape layer. Excludes objects on other layers, 
                            including the default layer
        """
        forms = []
        layer_names = _rs.LayerNames()
        select = True
        for name in layer_names:
            if _l.Layer.is_a_shape_layer(name):
                form = _rs.ObjectsByLayer(name, select)
                forms.append(form)
        return forms

    def highlight_as_type(self, form_type):     ##  from PyForm. Fix
        """Receives:
            form_type       str. 'a'| 'b' | 'c' | 'd'
        Applies color to its elements according to the form type. Displays 
        the reference circle (the circle around the reference point)? Returns:
            n               int. The number of elements modified if successful 
            None            if unsuccessful
        """
        if form_type == 'a':
            color = color_a
        elif form_type == 'b':
            color = color_b
        elif form_type == 'c':
            color = color_c
        elif form_type == 'd':
            color = color_d
        else:
            print("Need a legit form type: a, b, c, or d")
        n = self.colorize(color)
        # display ref point
        return n

    def colorize(self, color):
        """Receives:
            color           [num, num, num] 
        Changes the display color of the elements in the form. Returns:
            n               int. The number of elements modified
            None            if unsuccessful
        """
        if self.els == []:
            n = 0
        else:
            n = _rs.ObjectColor(self.els, color)
            _rs.Redraw()
        return n

    @classmethod
    def _get_text_dots_and_lines(cls, form):
        """Receives:
            form            (els, ref_point). May not be maximal
        Returns:
            text_dots       [guid, ...]. A list of text dot guids
            lines           [guid, ...]. A list of line guids
        """
        return (text_dots, lines)

    def enframe(self, frame_point, frame_diag, layer_name):
        """Adds a padded frame instance
        Receives:
            frame_point     point3d. The insertion point of the padded frame 
            frame_diag      point3d. The diagonal of the padded frame (from
                            the insertion point). diag_x > 0, diag_y > 0, 
                            diag_z > 0
            layer_name      str. The name of an existing layer
        Returns:
            frame_instance  guid. The guid of the padded frame instance. 
                            None if unsuccessful
        """
        if not (
            self.is_well_formed and
            frame_diag[0] > 0 and
            frame_diag[1] > 0 and
            frame_diag[2] > 0
        ):
            return_value = None
        else:
            frame_instance = _fr.Frame.new_instance(
                layer_name, frame_point, frame_diag)
            self.frame = frame_instance
            return_value = frame_instance
        return return_value

    def move_to(self, target):
        """Moves the elements and frame instance to the target. The 
        reference point is the insertion point of the frame instance
        Receives:
            target          point3d. The point to move to
        Returns:
            location        point3d. The final location of the frame instance. 
                            None if unsuccessful
        """
        if not self.is_well_formed:
            return_value = None
        else:
            framed_els = []
            framed_els.extend(self.els)
            framed_els.append(self.frame)
            frame_pos = self.get_frame_position()
            translation = _rs.PointSubtract(target, frame_pos)
            obs_moved = _rs.MoveObjects(framed_els, translation)
            if len(obs_moved) > 0:
                return_value = _rs.BlockInstanceInsertPoint(self.frame)
            else:
                return_value = None
        return return_value

    def move_by(self, translation):             ##  2018-10-04 11:06
        """Receives:
            translation     (num, num, num). The translation
        Moves self by the translation. Returns:
            value           boolean. True if successful. False otherwise
        """
        pass

    def select(self):
        """Selects the elements and frame
        Returns:
            n_obs           int. The number of objects (i.e., elements and 
                            frame) selected
        """
        n_obs = _rs.SelectObjects(self.els)
        if self.frame:
            n_frame = _rs.SelectObject(self.frame)
            n_obs += n_frame
        return n_obs

    def get_frame_position(self):
        """Finds the lower left front corner of the frame. Returns:
            pos             point3d. None if unsuccessful
        """
        if not self.is_well_formed:
            return_value = None
        elif not self.frame:
            return_value = None
        else:
            return_value = _rs.BlockInstanceInsertPoint(self.frame)
        return return_value

    def __hash__(self):
        els_hash = hash(self.els)
        point_hash = hash(self.ref_point)
        value = hash(els_hash, point_hash)
        return value

    def __eq__(self, other):
        value = False
        if (
            self.els == other.els and
            self.ref_point == other.ref_point
        ):
            value = True
        return value

    def __ne__(self, other):
        value = False
        if (
            self.els != other.els or
            self.ref_point != other.ref_point
        ):
            value = True
        return value

    # def __str__(self):
        # pass

    # def __repr__(self):
    #     pass

    @classmethod
    def _print_error_message(cls, method_name, message):
        print '%s.%s:\n    %s' % (cls.__name__, method_name, message)






