from view import arrow as               _a
from view import element as             _e
from view import frame as               _fr
from view import rh_form as             _rf
import rhinoscriptsyntax as             _rs
from view import settings as            _s
from view import view_main as           _vm
from view import view_utilities as      _vu

class RhRule(object):
    """Contains:
        a                   RhForm. The left form. Must not be empty
        b                   RhForm. The right form. May be empty
    """
    color_a = _s.Settings.color_left
    color_b = _s.Settings.color_right
    class_name = 'RhRule'

    def __init__(self, rh_form_a, rh_form_b, layer_name):
        """Receives:
            rh_form_a       RhForm. The left RhForm object. May be empty
            rh_form_b       RhForm. The right RhForm object. May be empty
            layer_name      str. The name of the layer containing the form 
                            elements. Guaranteed
        """
        class_name = RhRule.class_name
        method_name = '__init__'
        try:
            if not (
                type(rh_form_a) == _rf.RhForm and
                type(rh_form_b) == _rf.RhForm and
                type(layer_name) == str
            ):
                raise TypeError
            if not _rs.IsLayer(layer_name):
                raise ValueError
        except TypeError:
            message = "%s %s" % (
                "The arguments must be two RhForms", 
                "and a valid layer name")
            _vu._print_error_message(class_name, method_name, message)
            self.is_well_formed = False
        except ValueError:
            message = "The layer name must exist"
            _vu._print_error_message(class_name, method_name, message)
            self.is_well_formed = False
        else:
            self.a = rh_form_a
            self.b = rh_form_b
            self.layer_name = layer_name
            self.cbpp = self.common_bounding_points = (
                RhRule._find_common_bounding_points(self.a, self.b))
            frame_point, frame_diag = self._find_padded_frame_specs(self.cbpp)
            self.a.enframe(frame_point, frame_diag, self.layer_name)
            self.b.enframe(frame_point, frame_diag, self.layer_name)
            self.frame_diag = frame_diag
            self.arrow = None
            self.is_well_formed = True

    @classmethod
    def _find_common_bounding_points(cls, rh_form_a, rh_form_b):
        """Finds the common bounding points of the left and right forms. 
        Receives:
            rh_form_a       RhForm. The left rh_form
            rh_form_b       RhForm. The right rh_form
        Returns:
            cbpp            [point3d, ...]. A list of the 8 point3d objects 
                            of the bounding box of the two forms. Empty if 
                            both rh_forms are empty. None if unsuccessful
        """
        els_a = rh_form_a.els
        els_b = rh_form_b.els
        combined_els = []
        combined_els.extend(els_a)
        combined_els.extend(els_b)
        if combined_els == []:
            cbpp = []
        else:
            cbpp = _rs.BoundingBox(combined_els)
        return cbpp

    @classmethod
    def _find_padded_frame_specs(cls, bounding_points):
        """Finds the position (insertion point) and diagonal of the padded 
        frame. 
        Receives:
            bounding_points [point3d, ...]. A list of 8 point3ds. Empty for 
                            an empty shape
        Returns:
            frame_point     point3d. The front lower left point of the frame
            frame_diag      point3d. The diagonal from the insertion point
        """
        bpp = bounding_points
        if bpp == []:
            frame_pad_min = _s.Settings.frame_pad_minimum
            origin_p3d = _s.Settings.origin_p3d
            frame_point = _rs.PointSubtract(
                origin_p3d, 
                (frame_pad_min, frame_pad_min, 0))
            diag_x = frame_pad_min * 2
            diag_y = frame_pad_min * 2
            diag_z = frame_pad_min
            xyz_diag = (diag_x, diag_y, diag_z)
            frame_diag = _vu.get_point3d(xyz_diag)
        else:

            bpp_diag = _rs.PointSubtract(bpp[6], bpp[0])
            bpp_x, bpp_y, bpp_z = tuple(bpp_diag)
            pad_x = _fr.Frame.find_pad(bpp_x)
            pad_y = _fr.Frame.find_pad(bpp_y)
            pad_z = _fr.Frame.find_pad(bpp_z)

            bpp_0x, bpp_0y, bpp_0z = tuple(bpp[0])
            frame_px = bpp_0x - pad_x
            frame_py = bpp_0y - pad_y
            frame_pz = bpp_0z - 0
            frame_point = _vu.get_point3d((
                frame_px, 
                frame_py, 
                frame_pz))

            frame_dx = bpp_x + pad_x + pad_x
            frame_dy = bpp_y + pad_y + pad_y
            frame_dz = bpp_z +     0 + pad_z
            frame_diag = _vu.get_point3d((
                frame_dx, 
                frame_dy, 
                frame_dz))
        specs = (frame_point, frame_diag)
        return specs

    @classmethod
    def new_from_object_lists(cls, obs_a, obs_b, layer_name):
        """Creates a RhRule object from two lists of objects
        Receives:
            obs_a           [guid, ...]. A list of object guids in the left 
                            shape
            obs_b           [guid, ...]. A list of object guids in the right 
                            shape
            layer_name      str. The name of the layer containing the objects
        Returns:
            rh_rule         RhRule. The RhRule object specified by the lists. 
                            None if unsuccessful
        """
        rh_forms = []
        for obs_list in [obs_a, obs_b]:
            rh_form = _rf.RhForm(obs_list)
            if not rh_form.is_well_formed:
                return None
            else:
                rh_form = _rf.RhForm(obs_list)
                rh_forms.append(rh_form)
        rh_form_a, rh_form_b = rh_forms
        return_value = rh_rule = RhRule(rh_form_a, rh_form_b, layer_name)
        return return_value

    @classmethod
    def is_a_rule(cls, obs):                    ##  Used by ViewMain
        """Determines whether a list of objects makes up a rule.
        Receives:
            obs             [guid, ...]. A list of ob guids on the same 
                            layer. Guaranteed by the calling function
        Returns:
            value           boolean. True if the objects contain exactly 2 
                            non-overlapping frames. False otherwise
        """
        value = False
        if obs == []:
            pass
        else:
            group_name = layer_name = _rs.ObjectLayer(obs[0])
            frames = _fr.Frame.extract_instances(obs)
            n_frames = len(frames)
            if n_frames == 2:
                fr1, fr2 = frames[0:2]
                if (_rs.IsGroup(group_name) and
                    _rs.IsObjectInGroup(fr1, group_name) and
                    _rs.IsObjectInGroup(fr2, group_name) and
                    not _fr.Frame.overlaps_object(fr1, fr2)
                ):
                    value = True
        return value

    @classmethod
    def get_objects(cls):
        """Returns:
            all_guids       [guid, ...]. A list of guids of Rhino objects on 
                            rule layers
        """
        layer_names = _vm.ViewMain.get_rule_layer_names()
        all_guids = []
        for name in layer_names:
            guids_by_layer = _rs.ObjectsByLayer(name)
            all_guids.extend(guids_by_layer)
        return all_guids

    def find_frame_dy(self):
        """Returns:
            dy              num. The y-size of the frame. None if unsuccessful
        """
        return_value = None
        if self.is_well_formed:
            return_value = dy = self.frame_diag[1]
        return return_value

    def arrange(self):
        """Moves the left and right rh_forms to the display area. Draws the 
        arrow in the display area. Groups the elements, the frames, and the 
        arrow; the name is the rule_name. Returns: 
            frame_b_pos     point3d. The location of the lower left corner 
                            of the padded right rh_form. None if unsuccessful
        """
        if not self.is_well_formed:
            return_value = None
        else:
            frame_a_pos, arrow_pos, frame_b_pos = (
                self._find_rule_part_positions())
            value_a = self.a.move_to(frame_a_pos)
            value_b = self.b.move_to(frame_b_pos)
            self.arrow = _a.Arrow.new_instance(self.layer_name, arrow_pos)
            if (value_a and
                value_b and
                self.arrow
            ):
                _rs.AddGroup(self.layer_name)
                n_a_els = _rs.AddObjectsToGroup(self.a.els, self.layer_name)
                n_a_fr = _rs.AddObjectsToGroup(self.a.frame, self.layer_name)
                n_b_els = _rs.AddObjectsToGroup(self.b.els, self.layer_name)
                n_b_fr = _rs.AddObjectsToGroup(self.b.frame, self.layer_name)
                n_arrow = _rs.AddObjectsToGroup(self.arrow, self.layer_name)
                return_value = self.b.get_frame_position()
            else:
                return_value = None
        return return_value

    def _find_rule_part_positions(self):
        """Finds the positions of the left shape, arrow, and right shape. 
        Returns:
            a_pos           point3d. The lower left corner of the left form 
                            frame
            arrow_pos       point3d. The center of the arrow instance
            b_pos           point3d. The lower left corner of the right form 
                            frame
            None            if unsuccessful
        """
        if not self.is_well_formed:
            return_value = None
        else:

            (kx, ky, kz) = k = _s.Settings.rule_origin_offset_factor
            frame_dx, frame_dy, frame_dz = self.frame_diag
            lr_x = frame_dx * kx
            lr_y = frame_dy * ky
            lr_z = frame_dz * kz
            lr_pos = lower_right_position = (lr_x, lr_y, lr_z)
            part_vx = part_x_vector = _vu.get_point3d(
                (self.frame_diag[0], 0, 0))
            a_pos = _rs.PointSubtract(lr_pos, (part_vx * 3))
            arrow_dx = lr_pos[0] - (self.frame_diag[0] * 1.5)
            arrow_dy = lr_pos[1] + (self.frame_diag[1] * 0.5)
            arrow_pos = _vu.get_point3d((arrow_dx, arrow_dy, 0))
            b_pos = _rs.PointSubtract(lr_pos, (part_vx * 1))
            return_value = (a_pos, arrow_pos, b_pos)
        return return_value

    def select(self):
        self.a.select()
        self.b.select()
        _rs.SelectObject(self.arrow)



