import                                  copy
from view import element as             _e
from view import frame as               _fr
from view import grammar as             _g
from view import layer as               _l
from view import py_form as             _pf
from view import py_rule as             _pr
from view import rh_form as             _rf
from view import rh_rule as             _rr
import rhinoscriptsyntax as             _rs
from view import settings as            _s
from view import text_dot as            _td

block_type = _s.Settings.block_type
origin_object = _rs.AddPoint(0, 0, 0)
origin = _rs.PointCoordinates(origin_object)
origin_p3d = _s.Settings.origin_p3d

"""
Terms:
    An object is a Rhino object.
    An element is a Rhino object that represents a shape grammar basic 
        element, i.e., currently text dot, line. Later: curve, plane, 
        surface, annotation? 
    A form is a set of Rhino elements, i.e., a Rhino representation of a 
        shape. It may not be maximal. A reference point is specified by the 
        user to show the form's position relative to another form (in a rule). 
        The default is the global origin. 

Rhino object user-definable attributes:
    ObjectColor:        (display) color     weight
    ObjectPrintColor:   print color
    ObjectPrintWidth:   print width
    ObjectName:         name                label
"""

class ViewMain(object):
    """
    request (from user)
    draw (to document)
    """

    def __init__(self):
        _l.Layer.make_default()

    ### request

    def request_form_and_subform(self):
        """Precondition:
            The grammar must not be empty. 
        Prompts the user to select the current form and the current subform. 
        Returns:
            py_c            PyForm. The current py_form. Non-empty
            py_d            PyForm. The current sub_py_form. Non-empty
            None            if cancelled or unsuccessful
        """
        cancelled = None
        if _g.Grammar.is_empty():
            message = "The document is empty. Nothing to select"
            print(message)
            return_value = None
        else:
            message_c = 'Select the current shape'
            py_c = current_py_form = self.request_non_empty_form(message_c)
            if py_c:
                message_d = 'Select the current subshape'
                py_d = self._request_form(message_d)
                if py_d:
                    if py_d.is_empty:
                        py_d = py_c
                    return_value = (py_c, py_d)
                else:
                    return_value = cancelled
            else:
                return_value = cancelled
        return return_value

    def request_new_rule(self):
        """Prompts the user to specify a new rule. Returns:
            new_py_rule     PyRule. None if unsuccessful
        """
        message_a = 'Select the left rule shape'
        message_b = 'Select the right rule shape'
        pf_a = self._request_form_with_ref_point(message_a)
        pf_b = self._request_form_with_ref_point(message_b)
        new_py_rule = _pr.PyRule(pf_a, pf_b)
        if not new_py_rule.is_well_formed:
            return_value = None
        else:
            return_value = new_py_rule
        return return_value

    def _request_confirmation_duplicate_rule(self):
        """Requests user to confirm creation of duplicate rule. Returns:
            value           boolean
        """
        message = "This rule already exists. Create a second copy? [y or n]"
        response = _rs.GetString(message)
        yes = ['Y', 'y', 'Yes', 'yes']
        no = ['N', 'n', 'No', 'no']
        yes_no = copy.copy(yes)
        yes_no.extend(no)
        while response not in yes_no:
            response = _rs.GetString(message)
        if response in yes:
            value = True
        elif response in no:
            value = False
        else:
            pass
        return value

    def request_rule(self):
        """Prompts the user to select an existing rule. Returns:
            py_rule_id_str  str. None if cancelled or unsuccessful
        """
        message = "%s %s" % (
            "Select an existing rule.", 
            "Then you may have to wait a little"
        )
        method_name = 'request_rule'
        message_obs = message
        obs = self._request_objects(message_obs)
        while obs == []:
            title = method_name
            message = "Nothing was selected"
            response = self._request_retry_or_cancel(title, message)
            if response == 'retry':
                obs = self._request_objects(message_obs)
            elif response == 'cancel':
                return_value = cancelled = None
                break
            else:
                print("%s: We shouldn't have gotten here" % method_name)
                return_value = cancelled = None
                break
        if obs == None:
            return_value = cancelled = None
        # elif obs_on_several_layers            ##  new case!
        #   pass
        elif obs == []:
            return_value = cancelled = None
        elif _rr.RhRule.is_a_rule(obs):
            py_rule_id_str = layer_name = _rs.ObjectLayer(obs[0])
            return_value = py_rule_id_str
        else:
            return_value = trouble = None
        return return_value

    def _request_form(self, message):
        """Receives:
            message         str
        Prompts the user to select 0 or more objects (i.e., the current 
        subshape D). Returns:
            py_form         PyForm. Possibly empty. It has the origin as ref 
                            point
        """
        obs = self._request_objects(message)
        ref_point = origin_p3d
        form = _rf.RhForm(obs, ref_point)
        py_form = _pf.PyForm.new_from_objects_and_ref_point(obs, ref_point)
        return py_form

    def _request_form_with_ref_point(self, message):
        """Receives: 
            message         str
        Prompts the user to select 0 or more objects and the reference point 
        (i.e., the right rule shape B). Returns:
            py_form         PyForm. Possibly empty
        """
        obs = self._request_objects(message)
        if obs == []:
            ref_point = origin_p3d
        else:
            message_ref_point = "Select the reference point"
            ref_point = self._request_point(message_ref_point)
        form = _rf.RhForm(obs, ref_point)
        py_form = _pf.PyForm.new_from_objects_and_ref_point(obs, ref_point)
        return py_form

    def request_non_empty_form(self, message):
        """Receives:
            message         str
        Prompts the user to select at least one object (i.e., the current 
        shape C). Colors the objects. Returns:
            py_form         PyForm. Non-empty with the origin as ref point
            None            on cancel
        """
        method_name = 'request_non_empty_form'
        obs = self._request_objects(message)
        while obs == []:
            title = method_name
            message = 'The selection must not be empty'
            response = self._request_retry_or_cancel(title, message)
            if response == 'retry':
                obs = self._request_objects(message)
            elif response == 'cancel':
                obs = cancelled = None
                break
            else:
                print("%s: we shouldn't be here" % method_name)
                break
        if not obs:
            return_value = cancelled
        else:
            ref_point = origin
            form = _rf.RhForm(obs, ref_point)
            py_form = _pf.PyForm.new_from_objects_and_ref_point(
                obs, ref_point)
            return_value = py_form
        return return_value

    def request_non_empty_form_with_ref_point(self, message):
        """Receives:
            message         str
        Prompts the user to select at least one object and the reference 
        point (i.e., the left rule shape A). Returns:
            py_form         PyForm. Non-empty
            None            on cancel
        """
        method_name = 'request_non_empty_form_with_ref_point'
        obs = self._request_objects(message)
        while obs == []:
            title = method_name
            message = "The shape must not be empty"
            response = self._request_retry_or_cancel(title, message)
            if response == 'retry':
                obs = self._request_objects(message)
            elif response == 'cancel':
                return_value = cancelled = None
                break
            else:
                print("%s: We shouldn't have gotten here" % method_name)
                break
        else:
            message_ref_point = "Select the reference point"
            ref_point = self._request_point(message_ref_point)
            form = _rf.RhForm(obs, ref_point)
            py_form = _pf.PyForm.new_from_objects_and_ref_point(
                obs, ref_point)
            return_value = py_form
        return return_value

    def _request_objects(self, message):        ##  called
        """Receives:
            message         str
        Prompts the user to select objects (not frames). Returns:
            obs             [guid, ...]. A possibly empty list of object guids 
        """
        filter_ = 0
        obs = _rs.GetObjects(message, filter_)
        if not obs:
            obs = []
        return obs

    def _request_point(self, message):
        ##  Called by cm.get_rule_draw_next_shapes
        """Receives:
            message         str
        Prompts the user to select a point. Returns:
            point           point3d. The origin if no selection
        """
        value = _rs.GetPoint(message)
        if value == None:
            point = origin
        else:
            point = value
        return point

    def _request_retry_or_cancel(self, title, message):
        """Receives:
            title           str. The title of the message box
            message         str. The message to be displayed
        Returns:
            response_string str. 'retry' or 'cancel'
            response_number int. If unsuccessful
        """
        buttons = retry_and_cancel = 5
        response_number = _rs.MessageBox(message, buttons, title)
        if response_number == 4:
            return_value = response_string = 'retry'
        elif response_number == 2:
            return_value = response_string = 'cancel'
        else:
            return_value = response_number
        return return_value

    @classmethod
    def _extract_lines(cls, items):             ##  called by Draw
        """Receives:
            items           [guid, ...]. A list of guids of items
        Returns:
            lines           [guid, ...]. A list of guids of lines
        """
        lines = []
        for item in items:
            if (_rs.IsCurve(item) and
                _rs.IsLine(item)
            ):
                lines.append(item)
        return lines

    @classmethod
    def _extract_text_dots(cls, items):         ##  Called by Draw
        """Receives:
            items           [guid, ...]. A list of guids of items
        Returns:
            text_dots       [guid, ...]. A list of guids of text dots
        """
        text_dots = []
        for item in items:
            if _td.TextDot.is_text_dot(item):
                text_dots.append(item)
        return text_dots

    ### get

    @classmethod
    def get_rule_layer_names(cls):              ##  ControllerMain
        """Returns:
            rule_layer_names
                            [str, ...]. A list of names of layers that 
                            contain rules
        """
        rule_layer_names = []
        layer_names = _rs.LayerNames()
        for name in layer_names:
            if _l.Layer.contains_rule(name):
                rule_layer_name = name
                rule_layer_names.append(rule_layer_name)
        return rule_layer_names

    @classmethod
    def rh_rule_exists(cls, py_rule):
        """Receives:
            py_rule         PyRule. Guaranteed
        Returns:
            value           boolean. True if frames and an arrow exist on a 
                            layer with the same name root as the PyRule
        """
        value = False
        py_rule_id_str = str(py_rule.id_)
        rule_layer_names = cls.get_rule_layer_names()
        for name in rule_layer_names:
            if py_rule_id_str in name:
                value = True
                break
        return value

    ### draw
    
    def arrange_rh_rules(self, new_rule):
        """Receives:
            new_rule        RhRule. The newest RhRule object (drawn at the 
                            bottom)
        Arranges all rh_rules except new_rule by moving them in the positive 
        y direction. 

        # Arranges the rh_rules for display by moving their objects into the 
        # upper left quadrant. 

        Reinstates the current layer. Returns:
            n               int. The number of Rhino objects moved. None if 
                            unsuccessful
        """
        def _subtract_objects(obs_a, obs_b):
            obs_diff = []
            for ob_a in obs_a:
                if ob_a in obs_b:
                    ob_a_in_obs_b = ob_a
                else:
                    ob_a_not_in_obs_b = ob_a
                    obs_diff.append(ob_a_not_in_obs_b)
            return obs_diff

        return_value = None
        if new_rule:
            all_obs = _rr.RhRule.get_objects()
            new_obs = _rs.ObjectsByLayer(new_rule.layer_name)
            old_obs = _subtract_objects(all_obs, new_obs)
            rule_gap_factor = _s.Settings.rule_gap_factor
            dy = new_rule.find_frame_dy() * rule_gap_factor
            translation = (0, dy, 0)
            objects_moved = _rs.MoveObjects(old_obs, translation)
            return_value = len(objects_moved)
        return return_value

    def move_objects(self, comps_in, translation):
                                                ##  2018-09-20 09:30
        """Receives:
            comps_in        [comp, ...]. A list of rhino components, 
                            i.e., RhForms or RhRules
            trans           (num, num, num). The translation
        Moves the rhino components by the translation. Returns:
            comps_out       [comp, ...]. A list of the components moved. None 
                            if unsuccessful
        """
        return_value = None
        comps_out = []
        for comp in comps_in:
            if type(comp) in [RhForm, RhRule]:
                comp_out = comp.move(trans)
                if comp_out:
                    comps_out.append(comp_out)
            else:
                pass
            return_value = comps_out
        return return_value

    @classmethod
    def clear_all(cls):
        cls._clear_objects()
        cls._clear_blocks()
        cls._clear_layers()
        cls._clear_groups()
        cls._clear_data()
        # set Default layer                   ##  2018-10-04 10:49

    @classmethod
    def _clear_objects(cls):
        """Deletes all drawn objects, including locked objects, if any. 
        Returns:
            n_objects       int. The number of objects deleted, if successful
        """
        dummy_point = _rs.AddPoint(0, 0, 0)
        value = _rs.DeleteObject(dummy_point)
        include_lights = True
        include_grips = True
        objects = _rs.AllObjects(include_lights, include_grips)
        n_objects = _rs.DeleteObjects(objects)
        remaining_objects = _rs.AllObjects(include_lights, include_grips)
        _rs.UnlockObjects(remaining_objects)
        m_objects = _rs.DeleteObjects(remaining_objects)
        return n_objects + m_objects

    @classmethod                                ##  system-created blocks 
    def _clear_blocks(cls):                     ##  only?
        block_names = _rs.BlockNames()
        for name in block_names:
            _rs.DeleteBlock(name)

    @classmethod
    def _clear_layers(cls):
        """Deletes all [user-created?] layers. Leaves Default layer
        """
        default_layer_name = _s.Settings.default_layer_name
        layer_names = _rs.LayerNames()
        if not default_layer_name in layer_names:
            _rs.AddLayer(default_layer_name)
        _rs.CurrentLayer(default_layer_name)
        for layer_name in layer_names:
            if not layer_name == default_layer_name:
                value = _rs.DeleteLayer(layer_name)
                # print('Cleared layer %s: %s' % (layer_name, value))

    @classmethod
    def _clear_groups(cls):
        """Deletes all groups 
        """
        groups = _rs.GroupNames()
        if groups:
            for group in groups:
                _rs.DeleteGroup(group)

    @classmethod
    def _clear_data(cls):
        _rs.DeleteDocumentData()


